#!/usr/bin/env node

/** This file is part of Domotz Agent.
 * Copyright (C) 2017  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

var mod_path = require("path");
var mod_fs, mod_http, mod_https, mod_child_process;

if (global.jasmine) {
    mod_fs = {};
    mod_http = {};
    mod_https = {};
    mod_child_process = {};
} else {
    mod_fs = require("fs");
    mod_http = require("http");
    mod_https = require("https");
    mod_child_process = require("child_process");
}

console.debug = function () {
};

const MODULE_NAME = "AGENT_UPDATER";
const BASE_WARMUP_PERIOD = 5 * 1000;  // 5 seconds
const ONE_HOUR_IN_MSEC = 60 * 60 * 1000;  // 1 hour
const UPDATE_CHECK_PERIOD = 11 * 60 * 60 * 1000;  // 11 hours
const MAX_PROCESS_AGE = 48 * 60 * 60 * 1000;  // 2 days
const UPDATE_INFO_HOSTNAME = "portal.domotz.com";
const UPDATE_INFO_PATH = "/assets/domotz-npm-agent.json";
const NODE_COMMAND = mod_path.join(process.env.DOMOTZ_BIN_DIR, "domotz_node");
const NPM_CACHE_DIR = process.env.DOMOTZ_CACHE_DIR || process.env.DOMOTZ_ROOT_DIR;
const NPM_BASE_COMMAND = [mod_path.join(process.env.DOMOTZ_BIN_DIR, "domotz_npm"), "--cache",
    mod_path.join(NPM_CACHE_DIR, ".npm"), "--userconfig", mod_path.join(NPM_CACHE_DIR, ".npmrc"), "--unsafe-perm"];

"use strict";

var pkgJson = require(mod_path.join(process.env.DOMOTZ_LIB_DIR, 'node_modules', 'domotz-remote-pawn', 'package.json'));
var envCopy = Object.create(process.env);
envCopy.HOME = '/tmp';
var spawnOptions = {stdio: 'ignore', shell: false, env: envCopy};
var updaterStartTimestamp;

var updaterStatus = {
    error: false,
    update_in_progress: false,

    modifyCurrentStatus: function (in_progress, error) {
        this.update_in_progress = in_progress;
        this.error = error;
    },
    startProgress: function () {
        if (this.update_in_progress === true) {
            return false;
        }
        this.modifyCurrentStatus(true, false);
        return true;
    }
};

// Derived from : https://stackoverflow.com/questions/18052762/remove-directory-which-is-not-empty
function recursiveRemoveFolder(pathname) {
    if (mod_fs.existsSync(pathname) && pathname !== '/') {
        mod_fs.readdirSync(pathname).forEach(function (file) {
            var currentPath = mod_path.join(pathname, file);
            if (mod_fs.lstatSync(currentPath).isDirectory()) {
                recursiveRemoveFolder(currentPath);
            } else {
                mod_fs.unlinkSync(currentPath);
            }
        });
        mod_fs.rmdirSync(pathname);
    }
}

function upgradePawn(registry, name, version, customFolderEraser) {
    var tmpInstallationFolder = mod_path.join(process.env.DOMOTZ_LIB_DIR, 'tmp_inst');
    var tmpPawnFolder = mod_path.join(tmpInstallationFolder, 'node_modules', 'domotz-remote-pawn');
    var stablePawnFolder = mod_path.join(process.env.DOMOTZ_LIB_DIR, 'node_modules', 'domotz-remote-pawn');
    var oldPawnFolder = stablePawnFolder + '_old';

    console.info("%s - Start Agent Update from %s. Name: %s, Version: %s (%s)", MODULE_NAME, registry, name, version,
        (new Date()).toISOString());

    var prepareFileSystemForInstallation = function () {
        console.info("%s - Cleaning up old temporary folders...", MODULE_NAME);
        customFolderEraser(tmpInstallationFolder);
        customFolderEraser(oldPawnFolder);
        console.info("%s - Creating new installation folders (%s)", MODULE_NAME, tmpInstallationFolder);
        mod_fs.mkdirSync(tmpInstallationFolder);
    };

    var postAgentInstallProcedure = function (resultOk) {
        console.info("%s - Post install procedure (result=%s)", resultOk);

        if (resultOk !== true) {
            console.error("%s - ERROR: remote pawn update failed", MODULE_NAME);
            updaterStatus.modifyCurrentStatus(false, true);
            process.exit(1);  // If the updater process is simulated, we can only exit at this point
        }

        try {
            if (process.env.DPLATFORM !== "docker") {
                mod_fs.renameSync(stablePawnFolder, oldPawnFolder);
                mod_fs.renameSync(tmpPawnFolder, stablePawnFolder);
            } else {
                mod_child_process.spawnSync("cp", ["-R", stablePawnFolder, oldPawnFolder], spawnOptions);
                mod_child_process.spawnSync("rm", ["-rf", stablePawnFolder], spawnOptions);
                mod_child_process.spawnSync("cp", ["-R", tmpPawnFolder, stablePawnFolder], spawnOptions);
                mod_child_process.spawnSync("rm", ["-rf", tmpPawnFolder], spawnOptions);
            }

            console.info("%s - Remote pawn updated", MODULE_NAME);
            updaterStatus.modifyCurrentStatus(false, false);
            customFolderEraser(tmpInstallationFolder);
        } catch (e) {
            console.error("%s - Error on post-upgrade procedure (%s)", MODULE_NAME, JSON.stringify(e));
            updaterStatus.modifyCurrentStatus(false, true);
        } finally {
            // Clean the npm cache and Force the agent restart
            customFolderEraser(mod_path.join(NPM_CACHE_DIR, ".npm"));
            var pid = parseInt(mod_fs.readFileSync(process.env.LISTENER_PID_FILE));
            console.info("%s - killing listener (pid=%d)", MODULE_NAME, pid);
            process.kill(pid, 'SIGTERM');
        }
    };

    prepareFileSystemForInstallation();

    var npm_command = NPM_BASE_COMMAND.concat(['--legacy-bundle', 'install', name + '@' + version, '--prefix',
        tmpInstallationFolder]);
    console.info("%s - Invoking npm installer |%s|", MODULE_NAME, JSON.stringify(npm_command));
    var npm_installer = mod_child_process.spawn(NODE_COMMAND, npm_command, spawnOptions);

    npm_installer.on('error', function (err) {
        console.error("%s - Error on npm install (%s). Cleaning up the npm cache.", MODULE_NAME, JSON.stringify(err));
        customFolderEraser(mod_path.join(NPM_CACHE_DIR, ".npm"));
        updaterStatus.modifyCurrentStatus(false, true);
    });

    npm_installer.on('exit', function (code) {
        postAgentInstallProcedure(code === 0)
    });
}

function handleJsonUpdateInfo(updateInfo, forceVersion, customUpgradeHandler) {
    var upgradeHandler = customUpgradeHandler || upgradePawn;
    var parsedUpdateInfo = JSON.parse(updateInfo.toString('utf8'));

    function tryobj(obj_var, keys_arr) {

        if (obj_var === null || obj_var === undefined) {
            return null;
        }

        var last_obj = obj_var;
        keys_arr.forEach( function(key) {

            if (last_obj[key] === null || last_obj[key] === undefined) {
                return null;
            }
            last_obj = last_obj[key];
        } );

        return last_obj;
    }

    var dplatform = process.env.DPLATFORM;
    var minReq_name  = tryobj(parsedUpdateInfo, [dplatform, "name"]) || tryobj(parsedUpdateInfo, ["name"]);
    var minReq_registry = tryobj(parsedUpdateInfo, [dplatform, "registry"]) || tryobj(parsedUpdateInfo, ["registry"]);
    var minReq_version = tryobj(parsedUpdateInfo, [dplatform, "version"]) || tryobj(parsedUpdateInfo, ["version"]);

    if (pkgJson.version < minReq_version || forceVersion === true) {
        upgradeHandler(minReq_registry, minReq_name, minReq_version, recursiveRemoveFolder);
        return;
    }
    updaterStatus.modifyCurrentStatus(false, false);
}

function getUpdateInfoViaHttps(jsonResponseCallback) {
    var requestOptions = {
        hostname: UPDATE_INFO_HOSTNAME,
        path: UPDATE_INFO_PATH + '?agent=' + encodeURIComponent(process.env.DPLATFORM + '.' + process.env.DARCHITECTURE +
            '.' + process.env.DVERSION + '.js.' + pkgJson.version),
        headers: {'Content-Type': 'application/json'}
    };

    console.debug('%s - getUpdateInfoViaHttps - path <%s>', MODULE_NAME, requestOptions.path);
    var handleRawHttpBody = function (res) {
        var body = new Buffer(0);

        res.on('data', function (chunk) {
            body = Buffer.concat([body, chunk])
        });
        res.on('end', function () {
            if (!this.complete || res.statusCode !== 200) {
                console.error("%s - HTTP Error (complete=%s, statusCode=%d)", this.complete, res.statusCode);
                updaterStatus.modifyCurrentStatus(false, true);
                return;
            }
            try {
                console.info("%s - Handling response from server (len=%d, timestamp=%s)", MODULE_NAME, body.length,
                    (new Date()).toISOString());
                jsonResponseCallback(body, false);
            } catch (e) {
                console.error("%s - Error on callback : %s", MODULE_NAME, JSON.stringify(e));
                updaterStatus.modifyCurrentStatus(false, true);
            }
        });
    };

    mod_https.get(requestOptions, handleRawHttpBody)
        .on('error', function (err) {
            console.error("Error: " + err.message);
            updaterStatus.modifyCurrentStatus(false, true);
        });
}

function taskScheduler() {
    console.debug("%s - Scheduler: checking for tasks to execute ...", MODULE_NAME);

    if (updaterStatus.startProgress() === true) {

        if ((new Date()).getTime() - updaterStartTimestamp.getTime() > MAX_PROCESS_AGE) {
            console.info("%s - Graceful shutdown");
            process.exit(0);
        }

        console.debug("%s - Scheduler: running getUpdateInfoViaHttps task", MODULE_NAME);
        setImmediate(function () {
            getUpdateInfoViaHttps(handleJsonUpdateInfo);
        });

    }
    setTimeout(taskScheduler, UPDATE_CHECK_PERIOD + Math.round(ONE_HOUR_IN_MSEC * Math.random()));
}

function createHttpService(unixSocket) {

    var requestHandler = function (request, response) {
        var body = new Buffer(0);

        response.on('error', function (err) {
            console.info("%s - Error sending response :%s", err)
        });

        if (request.method.toUpperCase() !== 'POST') {
            response.statusCode = 405;  // METHOD NOT ALLOWED
            response.end();
            return;
        }
        request.on('error', function (err) {
            console.info("%s - Error receiving request :%s", err)
        });
        request.on('data', function (chunk) {
            body = Buffer.concat([body, chunk])
        });
        request.on('end', function () {
            console.info("%s - Received request (len=%d, busy=%s, timestamp=%s)", MODULE_NAME, body.length,
                updaterStatus.update_in_progress, (new Date()).toISOString());
            if (updaterStatus.startProgress() === true) {
                setTimeout(function () {
                    try {
                        handleJsonUpdateInfo(body, true);
                    } catch (e) {
                        console.info("%s - Error serving request : %s", MODULE_NAME, JSON.stringify(e));
                        updaterStatus.modifyCurrentStatus(false, true);
                    }
                }, 1000);
            }
            response.statusCode = 202;  // ACCEPTED
            response.end();
        });
    };

    try {
        mod_fs.unlinkSync(unixSocket);
    } catch (e) {
        console.info("%s - Old Unix socket not found %s", MODULE_NAME, unixSocket);
    }
    mod_http.createServer(requestHandler).listen(unixSocket, function () {
        console.info("%s - Listening on unix socket %s", MODULE_NAME, unixSocket);
    }).on('error', function (err) {
        console.error("%s - Unable to listen on unix socket %s", MODULE_NAME, err);
    });
}

function start(customHttpService) {
    console.info("%s - Initializing the updater process (%s) ...", MODULE_NAME, (new Date()).toISOString());
    updaterStartTimestamp = new Date();

    if (!process.env.DOMOTZ_UPDATER_UNIX_SOCKET) {
        setTimeout(function () {
            console.error('Exit because of wrong environment variables')
        }, 30000);
        return;
    }
    var warmup = BASE_WARMUP_PERIOD + Math.round(4 * BASE_WARMUP_PERIOD * Math.random());
    console.info("%s - Starting the scheduler in %d seconds", MODULE_NAME, warmup / 1000);
    setTimeout(taskScheduler, warmup);
    console.info("%s - Starting the http service", MODULE_NAME);
    customHttpService(process.env.DOMOTZ_UPDATER_UNIX_SOCKET);
}


if (require.main === module) {
    start(createHttpService);
} else {
    if (global.jasmine) {
        module.exports.mock_mod_fs = mod_fs;
        module.exports.mock_mod_http = mod_http;
        module.exports.mock_mod_https = mod_https;
        module.exports.mock_mod_child_process = mod_child_process;
        module.exports.updaterStatus = updaterStatus;
        module.exports.taskScheduler = taskScheduler;
        module.exports.getUpdateInfoViaHttps = getUpdateInfoViaHttps;
        module.exports.handleJsonUpdateInfo = handleJsonUpdateInfo;
        module.exports.upgradePawn = upgradePawn;
    }
    module.exports.start = start;
    module.exports.createHttpService = createHttpService;
}
