function Get-ICS {
    [CmdletBinding()]
    param (
        [Parameter(Mandatory,
                   ValueFromPipeline,
                   ValueFromPipelineByPropertyName)]
        [Alias('Name')]
        [string[]]$InternetInterfaceName
    )
    BEGIN {
        regsvr32.exe /s hnetcfg.dll
        $netShare = New-Object -ComObject HNetCfg.HNetShare
    }
    PROCESS {
        foreach ($Interface in $InternetInterfaceName){
            $publicConnection = $netShare.EnumEveryConnection |
            Where-Object {
                $netShare.NetConnectionProps.Invoke($_).Name -eq $Interface
            }
            try {
                $Results = $netShare.INetSharingConfigurationForINetConnection.Invoke($publicConnection)
            }
            catch {
                Write-Warning -Message "An unexpected error has occurred for network adapter: '$Interface'"
                Continue
            }
            [pscustomobject]@{
                Name = $Interface
                SharingEnabled = $Results.SharingEnabled
                SharingConnectionType = $Results.SharingConnectionType
                InternetFirewallEnabled = $Results.InternetFirewallEnabled
            }
        }
    }
}

function Set-ICS {
    [CmdletBinding()]
    param (
        [Parameter(Mandatory,
                   ValueFromPipeline,
                   ValueFromPipelineByPropertyName)]
        [ValidateScript({
            If ((Get-NetAdapter -Name $_ -ErrorAction SilentlyContinue -OutVariable INetNIC) -and (($INetNIC).Status -ne 'Disabled' -or ($INetNIC).Status -ne 'Not Present')) {
                $True
            }
            else {
                Throw "$_ is either not a valid network adapter of it's currently disabled."
            }
        })]
        [Alias('Name')]
        [string]$InternetInterfaceName,
        [ValidateScript({
            If ((Get-NetAdapter -Name $_ -ErrorAction SilentlyContinue -OutVariable LocalNIC) -and (($LocalNIC).Status -ne 'Disabled' -or ($INetNIC).Status -ne 'Not Present')) {
                $True
            }
            else {
                Throw "$_ is either not a valid network adapter of it's currently disabled."
            }
        })]
        [string]$LocalInterfaceName,
        [Parameter(Mandatory)]
        [bool]$Enabled
    )
    BEGIN {
        if ((Get-NetAdapter | Get-ICS).SharingEnabled -contains $true -and $Enabled) {
            Write-Warning -Message 'Unable to continue due to Internet connection sharing already being enabled for one or more network adapters.'
            Break
        }
        regsvr32.exe /s hnetcfg.dll
        $netShare = New-Object -ComObject HNetCfg.HNetShare
    }
    PROCESS {
        $publicConnection = $netShare.EnumEveryConnection |
        Where-Object {
            $netShare.NetConnectionProps.Invoke($_).Name -eq $InternetInterfaceName
        }
        $publicConfig = $netShare.INetSharingConfigurationForINetConnection.Invoke($publicConnection)
        if ($PSBoundParameters.LocalInterfaceName) {
            $privateConnection = $netShare.EnumEveryConnection |
            Where-Object {
                $netShare.NetConnectionProps.Invoke($_).Name -eq $LocalInterfaceName
            }
            $privateConfig = $netShare.INetSharingConfigurationForINetConnection.Invoke($privateConnection)
        }
        if ($Enabled) {
            $publicConfig.EnableSharing(0)
            if ($PSBoundParameters.LocalInterfaceName) {
                $privateConfig.EnableSharing(1)
            }
        }
        else {
            $publicConfig.DisableSharing()
            if ($PSBoundParameters.LocalInterfaceName) {
                $privateConfig.DisableSharing()
            }
        }
    }
}

function firewallSetRules {
    Param
    (
        [Parameter(Mandatory=$true, Position=0)]
        [string] $LocalInterfaceName,
        [parameter(Mandatory=$true, Position=1)]
        [string] $Enable
    )

    if (-Not (Get-NetFirewallRule -DisplayName $LocalInterfaceName)) {
        netsh advfirewall firewall add rule name="$LocalInterfaceName" dir=in action=allow protocol=TCP profile=Private
        netsh advfirewall firewall add rule name="$LocalInterfaceName (echo - ICMPv4-In)" dir=in action=allow protocol=ICMPv4 profile=Private
    }

    netsh advfirewall firewall set rule name="$LocalInterfaceName" new enable=$Enable
    netsh advfirewall firewall set rule name="$LocalInterfaceName (echo - ICMPv4-In)" new enable=$Enable

}

$NIC = Get-NetRoute | where {$_.DestinationPrefix -eq '0.0.0.0/0'} | select -expand InterfaceAlias
$sleepPidFile = "C:\ProgramData\domotz\openvpn\sleep.pid"
$invocation = $MyInvocation.MyCommand.Name -join ' '

if ($args[0] -eq "-Start") {
    $CONNECTION_DURATION_SECONDS = $args[2]
    Start-Process "C:\Program Files (x86)\domotz\lib\OpenVPN\bin\domotz_openvpn.exe" -ArgumentList '--config', '"C:\ProgramData\domotz\openvpn\conf\server.conf"' -WindowStyle hidden -PassThru
    Start-Process -FilePath powershell.exe -ArgumentList ./$invocation, '-Set-ICS ', $args[1] -WindowStyle hidden
    $sleep = Start-Process -FilePath powershell.exe -ArgumentList "-Command Start-Sleep -s $CONNECTION_DURATION_SECONDS;", ./$invocation, '-Stop ', $args[1], ' -Force' -WindowStyle hidden -Passthru
    $sleep.Id | Set-Content $sleepPidFile
    Start-Service -Name RemoteAccess
    Start-Process -FilePath powershell.exe -ArgumentList ./$invocation, '-FirewallEnable ', $args[1] -WindowStyle hidden
    echo "CONNECTION DURATION SECONDS $CONNECTION_DURATION_SECONDS/s"
} elseif ($args[0] -eq "-Stop") {
    Start-Process -FilePath powershell.exe -ArgumentList ./$invocation, '-UnSet-ICS ', $args[1] -WindowStyle hidden
    Stop-Process -Name domotz_openvpn -Force
    Start-Process -FilePath powershell.exe -ArgumentList ./$invocation, '-FirewallDisable ', $args[1] -WindowStyle hidden
    Stop-Process -Id (Get-Content -Path $sleepPidFile) -Force
} elseif ($args[0] -eq "-Get") {
    Get-ICS -InternetInterfaceName "$NIC", $args[1]
} elseif ($args[0] -eq "-FirewallDisable") {
    firewallSetRules -LocalInterfaceName $args[1] -Enable 'no'
} elseif ($args[0] -eq "-FirewallEnable") {
    firewallSetRules -LocalInterfaceName $args[1] -Enable 'yes'
} elseif ($args[0] -eq "-Set-ICS") {
    Set-ICS -InternetInterfaceName "$NIC" -LocalInterfaceName $args[1] -Enable $true
    Set-NetIPInterface -InterfaceAlias $args[1] -DHCP Enable
    Set-NetConnectionProfile -InterfaceAlias "DomotzVPN" -NetworkCategory Private
} elseif ($args[0] -eq "-UnSet-ICS") {
    Set-ICS -InternetInterfaceName "$NIC" -LocalInterfaceName $args[1] -Enable $false
}
