/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Iacopo Papalini <ipapalini@domotz.com> on 10/09/15.
 */

const MODULE_NAME = 'SPEED_TEST';
const KILL_CHILD_TIMEOUT = 300000;
const INTERNAL_OOKLA_PLATFORMS = ['openwrt', 'luxul', 'raspberry'];

var lockSeed = Math.random();

var speedTest = function (message, resourceLocator) {

    const md5 = resourceLocator.utils.random.md5;
    const time = resourceLocator.utils.time;
    const mutex = resourceLocator.utils.mutex;
    const fork = resourceLocator.utils.proc.fork;
    const join = resourceLocator.path.join;
    const engineClient = resourceLocator.engineRequest;
    const ndtServers = resourceLocator.networkTools.ndtServers;
    const speedTest = resourceLocator.networkTools.speedTest;
    const mlabTest = resourceLocator.networkTools.mlabSpeedTest;
    const fast = resourceLocator.networkTools.fastSpeedTest;
    const ooklaServers = resourceLocator.networkTools.ooklaServers;
    var myConsole = resourceLocator.log.decorateLogs();

    const cid = message.correlationId || 'From API';
    const serverId = message.payload.server_id;
    const nonInteractive = message.non_interactive;

    var getSpeedTestSuite = resourceLocator.networkTools.speedTest.getSpeedTestSuite;
    var SPEED_TEST_SUITES = resourceLocator.networkTools.speedTest.SPEED_TEST_SUITES;

    var ooklaModulePath = join(__dirname, '..', '..', 'network_tools', 'ooklaSpeedTest.js');
    var ooklaCall = isInternalTest() ? callInternalOoklaTest : callExternalOoklaTest;

    var lockId = md5(lockSeed.toString());

    function lock () {
        return mutex.lock(MODULE_NAME, lockId);
    }

    function unlock () {
        mutex.unlock(MODULE_NAME, lockId);
    }

    var results = {
        stats: {
            start: new Date(),
            end: null,
            elapsed_milliseconds: null
        },
        speedTest: {}
    };

    function copyResults (ndtInfo) {
        myConsole.debug('NDT INFO %s', ndtInfo);

        results.stats.end = new Date();
        results.stats.elapsed_milliseconds = time.getDiff(results.stats);
        results.speedTest = speedTest.ndtInfoIntoDTO(ndtInfo);
        return speedTest.filterItems(ndtInfo);
    }

    function processOoklaResult (data, cb) {

        function sendResults (result) {
            var items = copyResults(result);
            cb(results);
            engineClient.sendWithPromise('/network/speed-test', 'POST', items);
        }

        if (data.blacklist && data.blacklist.length > 0) {
            myConsole.warn('(%s): Blacklist servers => %s', cid,
                JSON.stringify(data.blacklist));
            ooklaServers.blacklistServers(data.blacklist);
        }
        if (!data.error) {
            if (data.server) {
                /* store it as preferred server to perform the interactive speed test */
                ooklaServers.setServerAsPreferred(data.server);
            }
            myConsole.debug('(%s): Speed test success', cid);

            sendResults(data.result);
        } else {
            myConsole.debug('(%s): Speed test failure => %s', cid,
                data.error);
            cb(data.error, data.error);
        }
    }

    function callExternalOoklaTest (msg, cb) {
        var child = fork(ooklaModulePath, [], {silent: false});
        var killChildTimeout = setTimeout(function () {
            myConsole.info('(%s) killing child process by timeout', cid);
            child.kill();
            unlock();
        }, KILL_CHILD_TIMEOUT);

        child.send(msg);
        child.on('message', function (data) {
            clearTimeout(killChildTimeout);
            processOoklaResult(data, cb);
            child.kill();
            unlock();
        });
    }

    function callInternalOoklaTest (msg, cb) {
        var ooklaSpeedTest = require(ooklaModulePath);
        ooklaSpeedTest.main(msg, function (data) {
            processOoklaResult(data, cb);
            unlock();
        });
    }

    function ndtTest (callback) {
        ndtServers.findById(serverId)
            .then(speedTest.performSpeedTest)
            .then(copyResults)
            .then(function (items) {
                myConsole.debug('%s NDT CB with results %s', JSON.stringify(results));
                callback(results);
                myConsole.debug('%s NDT Send with promise %s', JSON.stringify(items));
                return engineClient.sendWithPromise('/network/speed-test', 'POST', items);
            })
            .catch(function (err) {
                callback(null, err);
            }).finally(function () {
            unlock();
        });
    }

    function interactiveOoklaTest (callback) {
        myConsole.info('(%s): interactive mode, searching for %s', cid, serverId);
        ooklaServers.findById(serverId)
            .then(function (server) {
                myConsole.info('(%s): found server: %s',
                    cid, JSON.stringify(server));
                var msg = {bestServer: server, serverList: null};
                ooklaCall(msg, callback);
            });
    }

    function nonInteractiveOoklaTest (callback) {
        myConsole.info('%s (%s): Not interactive mode, selecting N closest servers', cid);
        ooklaServers.getNearestServers()
            .then(function (serverList) {
                myConsole.info('(%s): Nearest Server: %s',
                    cid, JSON.stringify(serverList));
                var msg = {
                    bestServer: null,
                    serverList: serverList
                };
                ooklaCall(msg, callback);
            });
    }


    return {
        describe: function () {
            return cid + ' - ' + MODULE_NAME;
        },
        execute: function (callback) {
            if (!lock()) {
                var msg = MODULE_NAME + ' - (' + cid + '): Another Speed test is currently running';
                myConsole.warn(msg);
                return callback(null, {
                    status: 409,
                    message: msg
                });
            }

            var speedTestSuite = getSpeedTestSuite();

            if (speedTestSuite === SPEED_TEST_SUITES.NDT) {
                ndtTest(callback);
            } else if (speedTestSuite === SPEED_TEST_SUITES.BENCHMARK || speedTestSuite === SPEED_TEST_SUITES.MLAB) {
                myConsole.debug('%s selected Mlab as speed test suite', MODULE_NAME);
                mlabTest.performSpeedTest()
                    .then(mlabTest.formatOutput)
                    .then(copyResults)
                    .then(function (items) {
                        myConsole.debug('%s PRE_FAST RESULTS %s', JSON.stringify(items));
                        if (nonInteractive && getSpeedTestSuite() === SPEED_TEST_SUITES.BENCHMARK) {
                            myConsole.debug(' running fast.con benchmark test', MODULE_NAME);
                            fast.performSpeedTest().then(function (fastResult) {
                                // fast results are pushed in the rawdata section
                                items.rawdata = [JSON.stringify({fast_speed_test: fastResult})];
                                myConsole.debug('%s MLAB CB with results %s', JSON.stringify(results));
                                callback(results);
                                myConsole.debug('%s MLAB Send with promise %s', JSON.stringify(items));

                                return engineClient.sendWithPromise('/network/speed-test', 'POST', items);
                            });
                        } else {
                            myConsole.debug('%s MLAB CB with results %s', JSON.stringify(results));
                            callback(results);
                            myConsole.debug('%s MLAB Send with promise %s', JSON.stringify(items));
                            return engineClient.sendWithPromise('/network/speed-test', 'POST', items);
                        }
                    })
                    .catch(function (err) {
                        callback(null, err);
                    }).finally(function () {
                    unlock();
                });
            } else if (speedTestSuite === SPEED_TEST_SUITES.FAST) {
                fast.performSpeedTest()
                    .then(fast.formatOutput)
                    .then(copyResults)
                    .then(function (items) {
                        callback(results);
                        return engineClient.sendWithPromise('/network/speed-test', 'POST', items);
                    }).finally(unlock);
            } else if (speedTestSuite === SPEED_TEST_SUITES.OOKLA) {
                nonInteractive ? nonInteractiveOoklaTest(callback) : interactiveOoklaTest(callback);
            }
        },
        getSpeedTestSuite: getSpeedTestSuite
    };
};

module.exports.command = speedTest;


function isInternalTest () {
    var env = process.env;
    return INTERNAL_OOKLA_PLATFORMS.indexOf(env.DPLATFORM) !== -1 || env.DARCHITECTURE === 'x64';
}

