/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/

/**
 * This module wraps the nmap commands used for IP devices' interfaces binds (mac-ip-hostname)
 * Created by Iacopo Papalini <iacopo@domotz.com> on 18/05/16.
 */

module.exports.BindDiscoveryFactory = function (nmapCommands, nmapXml2Js, fastParser, resourceLocator) {
    "use strict";
    var log = resourceLocator.log;
    var q = resourceLocator.q;
    var lodash = resourceLocator.lodash;
    var proc = resourceLocator.utils.proc;
    var hostInfo = resourceLocator.discovery.host;
    var settingsManager = resourceLocator.settingsManager;

    var iteration = 0;

    var baseConsole = log.decorateLogs();
    var myConsole = baseConsole;
    var SPAWN_HARD_TIMEOUT = 180000;

    function setIteration(i) {
        iteration = i;
        myConsole = baseConsole.decorate(String(iteration));
    }

    function executeDiscoveryCommands(commandList) {
        myConsole.info('executeDiscoveryCommands START ');

        var promises = [];
        commandList.forEach(function (command) {
            promises.push(proc.exec(command, {timeout: SPAWN_HARD_TIMEOUT}).then(nmapXml2Js.toJSON).then(fastParser.parse));
        });
        return q.allSettled(promises);
    }

    function getFulfilledPromisesResults(results) {
        /**
         * see
         * https://github.com/kriskowal/q/wiki/API-Reference#promiseallsettled
         * for information about the structure returned by allSettled
         * This function returns only the device bindings for the discoveries that
         * actually succeeded.
         */
        var fulfilled = lodash.map(results, function (netScanResult) {
            if (netScanResult.state === 'fulfilled') {
                return netScanResult.value;
            }
        });
        return lodash.filter(fulfilled, function (res) {
            return res !== null;
        });
    }

    function getAgentHostname() {
        var tenant = process.env.TENANT;
        if (tenant === 'adi') {
            return 'Capture Box';
        }
        return 'Domotz';
    }

    /**
     * Discovers all the IP devices in the given interfaces. When done the callback is invoked with, as argument,
     * an Array of Object similar to:
     * {
     *   'mac': '[the mac address]',
     *   'ip': '[the ip address]',
     *   'host_name': '[the host name]'
     * }
     */
    function discoverAll() {
        var interfaces = hostInfo.getAllowedInterfacesList();
        var name;
        var commandList = [];
        var myInterfacesBindings = {};
        var broadCastDisabled = settingsManager.isBroadcastDiscoveryDisabled();
        var bigNetworkScanner = resourceLocator.discovery.bigNetworkScanner;

        myConsole.info('Discover ALL START - Interfaces: %s', JSON.stringify(interfaces));
        for (name in interfaces) {
            if (interfaces.hasOwnProperty(name)) {
                var iface = interfaces[name];
                if (!iface.ip) {
                    myConsole.warn('Discover ALL, missing ip address for interface: %s', name);
                    continue;
                }
                var ip = iface.ip;
                var mac = iface.mac;
                var netmask = iface.netmask;
                var nmapType = mac !== null ? nmapCommands.NMAP_TYPE.DEFAULT : nmapCommands.NMAP_TYPE.PING;
                var bigNetwork = bigNetworkScanner.isBigNetwork(iface);
                var enableNetworkScan = !broadCastDisabled && !bigNetwork;

                if (bigNetwork) {
                    myConsole.debug("Network %s (%s/%s) is a big network", name, ip, netmask);
                }

                if (nmapType === nmapCommands.NMAP_TYPE.PING) {
                    myConsole.debug('Custom Interface for Monitoring External Network: %s/%s', ip, netmask);
                } else {
                    var macUpperCase = mac.toUpperCase();
                    myConsole.debug('Real Interface for Monitoring Internal Network %s/%s. Add Binding: %s->%s', ip, netmask, mac, ip);
                    if (!myInterfacesBindings.hasOwnProperty(macUpperCase)) {
                        myInterfacesBindings[macUpperCase] = [];
                    }
                    myInterfacesBindings[macUpperCase].push({ip: ip, host_name: getAgentHostname()});
                }
                var commands;
                if (enableNetworkScan) {
                    if (mac !== null) {
                        myConsole.info("Scanning all layer 2 network: %s %s/%s with nmap type: %s", name, ip, netmask, nmapType);
                    } else {
                        myConsole.info("Scanning all layer 3 network: %s %s/%s with nmap type: %s", name, ip, netmask, nmapType);
                    }
                    commands = nmapCommands.getFastNmapCmd(ip, netmask, nmapType);
                    commandList = commandList.concat(commands);
                } else if (iface.ipList.length > 0) {
                    // When ip sweep scan is disabled, or we discover IPs slowly with a Big Network Scan,
                    // we must ping individually all the known IPs to detect if they are online
                    myConsole.info("Probing only known (%s) IP addresses on network %s %s/%s", iface.ipList.length, name, ip, netmask);
                    commandList = commandList.concat([nmapCommands.getRetryCmd(iface.ipList, nmapType)]);
                } else {
                    myConsole.info("Network scan disabled and no known IP addresses for network %s (%s/%s), doing nothing", name, ip, netmask);
                }
            }
        }
        return executeDiscoveryCommands(commandList).then(function (results) {
            myConsole.info('executeDiscoveryCommands END');

            var devices = getFulfilledPromisesResults(results);
            myConsole.info('Discover ALL END - Interfaces: %s', JSON.stringify(interfaces));

            devices.push(myInterfacesBindings);
            return devices;
        });

    }

    function discover(ipList, nmapType, hostTimeout) {
        myConsole.info('Discover RETRY called - IP addresses: %s', JSON.stringify(ipList));
        var commandList = [nmapCommands.getRetryCmd(ipList, nmapType, hostTimeout)];
        return executeDiscoveryCommands(commandList).then(function (results) {
            myConsole.info('executeDiscoveryCommands END ');
            myConsole.info('Discover RETRY END - IP addresses: %s', JSON.stringify(ipList));
            return getFulfilledPromisesResults(results);
        });
    }

    function assureInterfaces() {
        var deferred = q.defer();
        try {
            // We must not use getAllowedInterfacesList here, because it is possible to have an empty allowed list
            // as a valid state
            var interfaces = hostInfo.getInterfacesList();
            if (Object.keys(interfaces).length === 0) {
                myConsole.info('assureInterfaces - No interfaces available, trying to rediscover host info');
                hostInfo.recoverHostInfo().finally(
                    function () {
                        deferred.resolve();
                    });
            } else {
                deferred.resolve();
            }
        } catch(err) {
            myConsole.info('assureInterfaces - Error trying to get interfaces list %s', err);
            deferred.resolve();
        }
        return deferred.promise;
    }


    return {
        setIteration: setIteration,
        discoverAll: discoverAll,
        discover: discover,
        assureInterfaces: assureInterfaces
    };
};


