/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/

/**
 * This module performs retries on stale bindings
 * Created by Andrea Azzara <a.azzara@domotz.com> on 13/07/18.
 */
const RETRY_TRIALS = 5;
const RETRY_HOST_TIMEOUT = 10;

module.exports.factory = function (q, host, log, configuration, blacklist, interfacesBinding, async, lodash, bindDiscovery, nmapCommands) {
    "use strict";
    var iteration = 0;
    var baseConsole = log.decorateLogs();
    var myConsole = baseConsole;

    function setIteration(i) {
        iteration = i;
        myConsole = baseConsole.decorate(String(iteration));
    }

    setIteration(0);

    function getRetryInterval(conf) {
        var discovery = conf.discovery;
        return discovery && discovery.retry_interval ? discovery.retry_interval : 1000;
    }

    var retryInterval = getRetryInterval(configuration);


    function retryOnStaleBindingsLocalDevices() {
        return _retryOnStaleBindings(RETRY_TRIALS, false);
    }

    function retryOnStaleBindingsExternalDevices() {
        return _retryOnStaleBindings(RETRY_TRIALS, true);
    }

    function _getStaleDevices(externalDevices) {
        var staleList = interfacesBinding.listStale();

        var devicesToExclude = blacklist.getDeviceToExclude();

        var filteredStaleDevices = staleList.filter(function (e) {
            return devicesToExclude.indexOf(e.ip) < 0;
        });

        return lodash.filter(filteredStaleDevices, function (d) {
            var device = [d];
            if (externalDevices) {
                return host.deviceBelongsToExternalSubnet(device) || host.deviceBelongsToExternalHosts(device);
            } else {
                return !host.deviceBelongsToExternalSubnet(device) && !host.deviceBelongsToExternalHosts(device);
            }
        }).map(function (e) {
            return e.ip;
        });
    }


    function _retryOnStaleBindings(count, externalDevices) {
        myConsole.debug("retryOnStaleBindings START externalDevices=%s", externalDevices);

        var deferred = q.defer();

        function pingDevices(callback) {

            var staleDevices = _getStaleDevices(externalDevices);
            if (staleDevices.length === 0) {
                myConsole.debug("retryOnStaleBindings END , there are no stale devices");
                callback();
                return;
            }

            myConsole.info("retryOnStaleBindings, there are %s stale devices, retry attempts left %s", staleDevices.length, count);

            var nmapType = nmapCommands.NMAP_TYPE.DEFAULT;

            if (externalDevices) {
                nmapType = nmapCommands.NMAP_TYPE.PING;
            } else if (count % 2 !== 0) {
                nmapType = nmapCommands.NMAP_TYPE.DISABLE_ARP_PING;
            }

            count--;

            bindDiscovery.discover(staleDevices, nmapType, RETRY_HOST_TIMEOUT)
                .then(interfacesBinding.setFreshDiscoveredBindings).then(function () {
                if (count === 0) {
                    myConsole.info("Exiting retry loop, devices still stale: %d", staleDevices.length);
                    callback();
                } else {
                    callback('continue');
                }
            });
        }

        async.retry({times: count, interval: retryInterval}, pingDevices, function () {
            myConsole.debug("retryOnStaleBindings END externalDevices=%s", externalDevices);

            deferred.resolve();
        });

        return deferred.promise;
    }


    return {
        retryOnStaleBindingsLocalDevices: retryOnStaleBindingsLocalDevices,
        retryOnStaleBindingsExternalDevices: retryOnStaleBindingsExternalDevices,
        _getStaleDevices: _getStaleDevices,
        _retryOnStaleBindings: _retryOnStaleBindings,
        setIteration: setIteration
    };
};


