/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * This module has the responsibility to send the discovered IP devices to the engine, choosing between a full or
 * a delta report following an internal logic
 *
 * Created by Iacopo Papalini <iacopo@domotz.com> on 19/05/16.
 */


var engineIpDiscoveryReportStrategyFactory = function (resourceLocator) {
    'use strict';

    var interfacesBindingStorage = resourceLocator.interfacesBindingStorage;
    var host = resourceLocator.discovery.host;
    var discoveryCounter = resourceLocator.discovery.counter;
    var iteration = 0;

    var baseConsole = resourceLocator.log.decorateLogs();
    var myConsole = baseConsole;

    function setIteration (i) {
        iteration = i;
        myConsole = baseConsole.decorate(String(iteration));
    }

    function addAccumulatedDiscoveryData (keys, payload) {
        var discoveryData = resourceLocator.discovery.accumulator.fetch(keys);
        Object.keys(discoveryData).forEach(function (mac) {
            if (payload[mac] && discoveryData[mac]) {
                myConsole.info('Adding protocols %s to device %s', Object.keys(discoveryData[mac]), mac);
                payload[mac][0].data = discoveryData[mac];
            }
        });

        return payload;
    }

    return (function () {
        var ret = {
            setIteration: setIteration,
            reportDiscovery: function () {
                var seq = discoveryCounter.getNextSequenceNumber();

                var payload = discoveryCounter.isFullDiscovery() ? ret.createFullPayload() : ret.createDeltaPayload();

                /* We need to make sure we are not reporting status information on devices belonging to external
                 * hosts/subnets that are not currently configured (e.g., this may happen when a new configuration is
                 * retrieved from the server and a discovery is already in progress based on the old configuration). */
                payload = resourceLocator.lodash.omitBy(payload, function (dev, key) {
                    if (!dev) {
                        return false;
                    }
                    var omit = key.indexOf(interfacesBindingStorage.CUSTOM_NETWORK_PREFIX) === 0 &&
                        !host.deviceBelongsToExternalSubnet(dev) ||
                        key.indexOf(interfacesBindingStorage.EXTERNAL_HOST_PREFIX) === 0 && !host.deviceBelongsToExternalHosts(dev);
                    if (omit) {
                        myConsole.warn('External subnet, Skipping report for custom mac: ' + key);
                    }
                    return omit;
                });

                var deferred = resourceLocator.q.defer();
                if (isToSend(payload, seq)) {
                    var keys = Object.keys(payload);
                    payload = addAccumulatedDiscoveryData(keys, payload);
                    myConsole.debug('Sending IP Network device discovery report to engine. Sequence number: ' + seq);
                    resourceLocator.engineRequest.send('/device/ip?seq=' + seq, 'PUT', payload,
                        function () {
                            myConsole.info('Report successfully sent to engine. Sequence number: ' + seq);
                            discoveryCounter.incrementCounter();
                            resourceLocator.discovery.accumulator.flush(keys);
                            return deferred.resolve();
                        },
                        function (err) {
                            myConsole.warn('Error reporting IP network devices ' +
                                'to engine. Sequence number: ' + seq +
                                ' - Rolling back sequence number');
                            return deferred.reject(err);
                        });
                } else {
                    myConsole.info('No changes in IP network - no report will be sent to the agent');
                    discoveryCounter.incrementCounter();
                    deferred.resolve();
                }

                function isToSend (payload, seq) {
                    return seq === 0 || isNotEmpty(payload);
                }

                function isNotEmpty (payload) {
                    // http://stackoverflow.com/questions/679915/how-do-i-test-for-an-empty-javascript-object
                    return !(Object.keys(payload).length === 0 &&
                        payload.constructor === Object);
                }

                return deferred.promise;
            },
            createFullPayload: function () {
                return bindingListToDictionary(interfacesBindingStorage.listFresh());
            },
            createDeltaPayload: function () {
                var payload = {};
                var fresh = bindingListToDictionary(interfacesBindingStorage.listFresh());

                var addStaleInfo = function (stale) {
                    var mac = stale.mac;
                    if (fresh[mac]) {
                        if (payload[mac] === undefined) {
                            payload[mac] = [];
                        }
                        payload[mac] = payload[mac].concat(fresh[mac]);
                    }
                    if (payload[mac] !== undefined) {
                        return;
                    }
                    payload[mac] = null;
                };
                var addNewInfo = function (newOne) {
                    var mac = newOne.mac;
                    // new device is always fresh - add it with all the fresh IP addresses
                    payload[mac] = fresh[mac];
                };

                interfacesBindingStorage.listStale().forEach(addStaleInfo);
                interfacesBindingStorage.listNew().forEach(addNewInfo);

                return payload;


            }
        };

        function bindingListToDictionary (list) {
            var ret = {};
            list.forEach(function (binding) {
                if (ret[binding.mac] === undefined) {
                    ret[binding.mac] = [];
                }
                ret[binding.mac].push({ip: binding.ip, host_name: binding.hostname});
            });
            return ret;
        }

        return ret;
    }());


};

module.exports.engineIpDiscoveryReportStrategyFactory = engineIpDiscoveryReportStrategyFactory;
