/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Tommaso Latini <tommaso@domotz.com> on 27/08/15.
 */


const IPV4_BITS = 32;
const DOMOTZ_NETMASK = 24;
const SCANNED_IP_PER_NMAP = Math.pow(2, IPV4_BITS - DOMOTZ_NETMASK);

var NMAP_TYPE = {
    DEFAULT: 'default',
    DISABLE_ARP_PING: 'disable_arp_ping',
    PING: 'ping'
};

var _ = require('lodash');

function getPortString(discovery) {
    return _.sortBy(discovery.ports).join(',');
}

function getPortUDPString(discovery) {
    return _.sortBy(discovery.udp_ports).join(',');
}

var factory = function (ip, configuration, resourceLocator) {
    var myConsole = resourceLocator.log.decorateLogs();

    var nmapBlackList = resourceLocator.nmapBlackList;
    var discoveryCounter = resourceLocator.discovery.counter;
    var settingsManager = resourceLocator.settingsManager;
    var platform = resourceLocator.utils.platform;

    function nmapCommandWithTimeout(hostTimeout) {
        hostTimeout = hostTimeout !== undefined ? hostTimeout : settingsManager.getNMAPDiscoveryTimeout();
        return 'domotz_nmap --host-timeout ' + hostTimeout + 's -oX - ';
    }
    var ipv4 = ip.Address4;

    var discovery = settingsManager.getDiscoverySettings();

    var portString = getPortString(discovery);
    var portUDPString = getPortUDPString(discovery);
    var maxNetmask = discovery.max_ip_netmask;

    var _getUDPCustomScriptDir = function () {
        var customScriptDir = '';
        if (settingsManager.getUDPScanSettings().use_nmap_custom_scripts) {
            var scriptDir = resourceLocator.path.join(__dirname, 'nmap_payloads');
            customScriptDir = ['--datadir', scriptDir].join(' ') + ' ';
        }
        return customScriptDir;
    };

    function getStandardNmapCmd(hostTimeout) {
        var ipList = nmapBlackList.getDeviceToExclude();
        var cmd = nmapCommandWithTimeout(hostTimeout) + (ipList.length ? '--exclude ' + ipList.join(',') + ' ' : ' ');
        if (platform.needsSudo()) {
            myConsole.debug("Platform %s needs sudo", process.env.DPLATFORM);
            cmd = 'sudo ' + cmd;
        }
        return cmd;
    }

    function getPingNmapCmd(networkList, nmapType, hostTimeout, disableDNS) {
        disableDNS = disableDNS !== undefined ? disableDNS : false;
        var baseCommand = getStandardNmapCmd(hostTimeout) + '-sP ';

        if (!discoveryCounter.isFullDiscovery() || disableDNS) {
            baseCommand += ' -n ';
        }

        if (!nmapType || nmapType === NMAP_TYPE.DEFAULT) {
            return baseCommand + networkList;
        } else if (nmapType === NMAP_TYPE.DISABLE_ARP_PING) {
            return baseCommand + ' --disable-arp-ping ' + networkList;
        } else if (nmapType === NMAP_TYPE.PING) {
            return baseCommand + '--disable-arp-ping -PE ' + networkList;
        } else {
            myConsole.error('getPingNmapCmd: Unrecognized nmap type');
        }
    }

    function getBaseLog(base, num) {
        return Math.log(num) / Math.log(base);
    }

    function getScanBigNetworksNmapCmd(networkAddress) {
        var params = settingsManager.getBigNetworksPerformanceParams();
        var timeout = params.timeout;
        var tValue = params.t_value;
        var subnetworkCount = params.subnetwork_count;

        var commands = [];
        var net = new ipv4(networkAddress);
        var baseIP = net.startAddress();
        var subnetMask = net.subnetMask;

        var ipCountPerNmap = Math.pow(2, IPV4_BITS - subnetMask) / subnetworkCount;

        var baseCommand = getStandardNmapCmd(timeout) + '-sn -n '; // -sn, same as -sP but not deprecated
        baseCommand += '-T' + tValue.toString() + ' '; // Makes scan faster (more aggressive)
        baseCommand = baseCommand.replace("-oX", "-oG"); // greppable output instead of xml

        for (var i = 0; i < subnetworkCount; i++) {
            var subNetwork = ipv4.fromInteger(parseInt(baseIP.bigInteger()) + i * ipCountPerNmap);
            var newSubnetMask = IPV4_BITS - getBaseLog(2, ipCountPerNmap);
            var debugNet = new ipv4(subNetwork.address + '/' + newSubnetMask);
            myConsole.verbose("%s %s %s", subNetwork.address + '/' + newSubnetMask,
                debugNet.startAddress().address,
                debugNet.endAddress().address
            );
            commands.push(baseCommand + subNetwork.address + '/' + newSubnetMask);
        }
        return commands;
    }

    function getFastNmapCmd(addr, netmask, nmapType) {
        if (netmask < maxNetmask) {
            myConsole.warn('Netmask Out of Range [%d-%d]: %d', maxNetmask, DOMOTZ_NETMASK, netmask);
            netmask = maxNetmask;
        }
        var commands = [],
            baseIP = new ipv4(addr + '/' + netmask).startAddress(),
            parallel = Math.pow(2, DOMOTZ_NETMASK - netmask);
        for (var i = 0; i < parallel; i++) {
            var network = ipv4.fromInteger(parseInt(baseIP.bigInteger()) + i * SCANNED_IP_PER_NMAP);
            var cidr = netmask > DOMOTZ_NETMASK ? netmask : DOMOTZ_NETMASK;
            var cmd = getPingNmapCmd(network.address + '/' + cidr, nmapType);
            commands.push(cmd);
        }
        myConsole.info("Using netmask %s for network %s (%s nmap commands)", netmask, addr, commands.length);
        return commands;
    }

    function getNETBIOSCmd() {
        var customScriptDir = _getUDPCustomScriptDir();
        return getStandardNmapCmd() + '-sU ' + customScriptDir + '--script nbstat.nse -p137 ';

    }

    function getRetryCmd(ipList, nmapType, hostTimeout) {
        return getPingNmapCmd(ipList.join(' '), nmapType, hostTimeout, true);
    }

    function getRetryArpDisabledCmd(ipList) {
        var targets = '';
        for (var i = 0; i < ipList.length; i++) {
            targets += ipList[i] + ' ';
        }
        return getPingNmapCmd(targets, NMAP_TYPE.DISABLE_ARP_PING);
    }

    function getPortScanType() {
        if (platform.isWindows()) {
            return '-sT';
        }
        return '-sS';
    }

    function getPortNmapCmd(ip) {
        // using TCP SYN scanning (https://nmap.org/book/man-port-scanning-techniques.html)
        return getStandardNmapCmd() + getPortScanType() + ' -p' + portString + ' ' + ip;
    }

    function getPortNmapCmdMultiIp(ips) {
        // using TCP SYN scanning (https://nmap.org/book/man-port-scanning-techniques.html)

        return getStandardNmapCmd() + getPortScanType() + ' -p' + portString + ' ' + ips.join(' ');
    }

    function getPortUDPNmapCmdMultiIp(ips) {
        var customScriptDir = _getUDPCustomScriptDir();
        var command = getStandardNmapCmd() + customScriptDir + '-n -sU --disable-arp-ping -p' + portUDPString + ' ' + ips.join(' ');
        myConsole.debug('udp_port_scanner nmap command %s', command);
        return command;
    }

    function getExplicitPortNmapCmd(ip, explicitPortString) {
        // using TCP SYN scanning (https://nmap.org/book/man-port-scanning-techniques.html)
        return getStandardNmapCmd() + '-n ' + getPortScanType() + ' -Pn --disable-arp-ping -p' + explicitPortString + ' ' + ip;
    }

    return {
        getPingNmapCmd: getPingNmapCmd,
        getRetryCmd: getRetryCmd,
        getRetryArpDisabledCmd: getRetryArpDisabledCmd,
        getFastNmapCmd: getFastNmapCmd,
        getPortNmapCmd: getPortNmapCmd,
        getPortNmapCmdMultiIp: getPortNmapCmdMultiIp,
        getPortUDPNmapCmdMultiIp: getPortUDPNmapCmdMultiIp,
        getExplicitPortNmapCmd: getExplicitPortNmapCmd,
        getNETBIOSCmd: getNETBIOSCmd,
        getScanBigNetworksNmapCmd: getScanBigNetworksNmapCmd,
        NMAP_TYPE: NMAP_TYPE
    };
};

module.exports.factory = factory;


