/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz UK LLP
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/
/**
 *
 * Created by Massimiliano Cuzzoli <massi@domotz.com> on 05/10/16.
 *
 */


const PROTOCOL_NAME = 'snmp';
const DISCOVERY_SEMAPHORE = 'SNMP_SEMAPHORE';
const OID_PORT_NUMBER = '1.3.6.1.2.1.2.1.0';
const snmpCommonDiscovery = require('./snmp_common');
const executorStrategy = require('./snmp_sequential_executor');
const TASK_DELAY = 20000;

var snmpDiscoveryFactory = function (resourceLocator, commandBuilder, responseRendering, snmpGetterFactory) {
    "use strict";
    var myConsole = resourceLocator.log.decorateLogs();
    var snmpInterfacesNumberStorage = resourceLocator.snmpInterfacesNumberStorage;
    var communitiesProvider = resourceLocator.snmpCommunitiesProvider;
    var interfacesBinding = resourceLocator.interfacesBindingStorage;
    var mutex = resourceLocator.utils.mutex;
    var q = resourceLocator.q;
    var platform = resourceLocator.utils.platform;
    var snmpDiscoverySnapshotData = {};
    var snmpCommon = snmpCommonDiscovery.snmpCommonOptions();

    var executor = platform.isSlow() ? executorStrategy.sequentialExecutor() : executorStrategy.parallelExecutor();

    var deferredSnmpBulkGet = function (snmpCommand, host, options, mac) {
        var deferred = q.defer();
        var commandOptions = snmpCommon.getSnmpOptions(options);
        myConsole.debug('Command Options are  %s -> %s', mac, JSON.stringify(commandOptions));

        snmpGetterFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                myConsole.debug('Received result for %s -> %s', mac, JSON.stringify(result));
                result.host = host;
                result.mac = mac;
                setTimeout(function () {
                    deferred.resolve(result);
                }, TASK_DELAY);
            },
            commandOptions
        );

        return deferred.promise;
    };

    var deferredSnmpBulkGetTask = function (snmpCommand, host, options, mac) {
        if (!platform.isSlow()) {
            return deferredSnmpBulkGet(snmpCommand, host, options, mac);
        } else {
            return function task() {
                return deferredSnmpBulkGet(snmpCommand, host, options, mac);
            };
        }
    };

    var lockDiscovery = function () {
        return mutex.lock(DISCOVERY_SEMAPHORE, 1);
    };

    var unlockDiscovery = function () {
        mutex.unlock(DISCOVERY_SEMAPHORE, 1);
    };

    var renderSnmpResults = function (results) {
        for (var i = 0; i < results.length; i++) {
            var host = results[i].host;
            var mac = results[i].mac;
            var data = null;
            try {
                data = responseRendering(resourceLocator, results[i]);
            } catch (e) {
                data = undefined;
            }
            if (host === undefined || mac === undefined || data === null) {
                myConsole.error("Something went wrong in rendering the SNMP result, skipping");

            } else {
                myConsole.debug('Parsed Result: %s - |%s|', mac, JSON.stringify(data));

                if (!data.error) {
                    snmpDiscoverySnapshotData[mac] = data;
                    if (data.oids && data.oids[OID_PORT_NUMBER]) {
                        var number = parseInt(data.oids[OID_PORT_NUMBER], 10);
                        snmpInterfacesNumberStorage.store(mac, number);
                    }
                }
            }
        }
    };

    var discover = function (callback, errCallback) {
        if (!lockDiscovery()) {
            myConsole.warn("WARNING: SNMP Discovery already in progress, ignoring the call");
            return q(null);
        }

        return communitiesProvider.get().then(function (communities) {
            var deferredResultList = [];
            myConsole.debug("Found communities %s ", JSON.stringify(communities));

            var oidsToDiscover = resourceLocator.settingsManager.getOidsToDiscover();
            interfacesBinding.listFresh().forEach(function (device) {
                var address = device.ip;
                var snmpCommand = commandBuilder.buildBulkGetCommand(oidsToDiscover, address);
                myConsole.debug("Command for device %s [mac %s]  %s ", device.ip, device.mac, JSON.stringify(snmpCommand));
                myConsole.debug("Communities for device %s [mac %s]  %s ", device.ip, device.mac, JSON.stringify(communities[device.mac]));
                deferredResultList.push(deferredSnmpBulkGetTask(snmpCommand, address, communities[device.mac], device.mac));
            });
            return executor.execute(deferredResultList, myConsole)
                .then(renderSnmpResults)
                .then(function () {
                    unlockDiscovery();
                    if (callback) {
                        callback(snmpDiscoverySnapshotData);
                    }
                });
        }).catch(function (err, err2) {
            unlockDiscovery();
            myConsole.error("Error on snmp information processing %s %s", err, err2);
            if (errCallback) {
                errCallback(err);
            }
        });
    };

    var applyFunctionToSnapshotEntries = function (functionToApply) {
        myConsole.debug("post-processing: " + JSON.stringify(snmpDiscoverySnapshotData));
        for (var macAddress in snmpDiscoverySnapshotData) {
            if (snmpDiscoverySnapshotData.hasOwnProperty(macAddress)) {
                functionToApply(macAddress, snmpDiscoverySnapshotData[macAddress], PROTOCOL_NAME);
            }
        }
    };

    var searchEntryIntoSnapshot = function (deviceAddress) {
        return snmpDiscoverySnapshotData[deviceAddress];
    };

    return {
        discover: discover,
        applyFunctionToSnapshotEntries: applyFunctionToSnapshotEntries,
        searchEntryIntoSnapshot: searchEntryIntoSnapshot,
        isDiscoveryInProgress: function () {
            return mutex.isLocked(DISCOVERY_SEMAPHORE);
        },
        getProtocolName: function () {
            return PROTOCOL_NAME;
        },
        getProtocolNameUpperCase: function () {
            return PROTOCOL_NAME.toUpperCase();
        },
        getOidForInterfaceNumber: function () {
            return OID_PORT_NUMBER;
        }
    };
};


module.exports.factory = snmpDiscoveryFactory;