/** This file is part of Domotz Agent.
 * Copyright (C) 2021 Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Created by fausto@domotz.com on 28/04/2021.
 *
 */
const DISCOVERY_SEMAPHORE = 'SNMP_CHECK_STATUS_SEMAPHORE';
const SYS_DESCR_OID = {'1.3.6.1.2.1.1.1.0': 'string'};
const snmpCommonDiscovery = require('./snmp_common');
const executorStrategy = require('./snmp_sequential_executor');
const mibDiscoveryEventManager = require('../../events/handler/snmpMibDiscovery');
const responseRendering = require('./response_processing').renderSnmpBasicData;
const TASK_DELAY = 1000;

var snmpCheckStatusDiscoveryFactory = function (resourceLocator, commandBuilder, snmpGetterFactory) {
    'use strict';
    var myConsole = resourceLocator.log.decorateLogs();
    var lodash = resourceLocator.lodash;
    var engineRequest = resourceLocator.engineRequest;
    var communitiesProvider = resourceLocator.snmpCommunitiesProvider;
    var interfacesBinding = resourceLocator.interfacesBindingStorage;
    var mutex = resourceLocator.utils.mutex;
    var q = resourceLocator.q;
    var platform = resourceLocator.utils.platform;
    var snmpCommon = snmpCommonDiscovery.snmpCommonOptions();

    var executor = platform.isSlow() ? executorStrategy.sequentialExecutor() : executorStrategy.parallelExecutor();

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * */
    /* WARNING: STATEFUL */
    var snmpDiscoverySnapshotData = {};
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * */

    function TaskBuilder () {}

    TaskBuilder.prototype.deferredSnmpBulkGet = function (snmpCommand, host, options, mac) {
        var deferred = q.defer();
        var commandOptions = snmpCommon.getSnmpOptions(options);
        myConsole.debug('Command Options are  %s -> %s', mac, JSON.stringify(commandOptions));

        snmpGetterFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                myConsole.debug('Received result for %s -> %s', mac, JSON.stringify(result));
                result.host = host;
                result.mac = mac;
                setTimeout(function () {
                    deferred.resolve(result);
                }, TASK_DELAY);
            },
            commandOptions
        );

        return deferred.promise;
    };

    TaskBuilder.prototype.deferredSnmpBulkGetTask = function (snmpCommand, host, options, mac) {
        if (!platform.isSlow()) {
            return this.deferredSnmpBulkGet(snmpCommand, host, options, mac);
        } else {
            var self = this;
            return function task () {
                return self.deferredSnmpBulkGet(snmpCommand, host, options, mac);
            };
        }
    };

    var lockDiscovery = function () {
        return mutex.lock(DISCOVERY_SEMAPHORE, 1);
    };

    var unlockDiscovery = function () {
        mutex.unlock(DISCOVERY_SEMAPHORE, 1);
    };

    function ResultHandler () {}

    ResultHandler.prototype.renderSnmpResults = function (results) {
        var snapshotData = {};
        for (var i = 0; i < results.length; i++) {
            var host = results[i].host;
            var mac = results[i].mac;
            var data = null;
            try {
                data = responseRendering(resourceLocator, results[i]);
            } catch (e) {
                data = undefined;
            }
            if (host === undefined || mac === undefined || data === null) {
                myConsole.error('Something went wrong in rendering the SNMP result, skipping');

            } else {
                myConsole.debug('Device result: %s - |%s|', mac, JSON.stringify(data));

                if (!data.error) {
                    snapshotData[mac] = {status: 'ON'};
                } else {
                    snapshotData[mac] = {status: 'OFF', error: data.description};
                }
            }
        }

        myConsole.debug('Total Results: |%s|', JSON.stringify(snapshotData));

        return snapshotData;
    };

    ResultHandler.prototype.sendResultToEngine = function (snmpDiscoveryData) {
        myConsole.info("Discovery terminated, uploading results ...");
        myConsole.info("Discovery terminated, Snapshot (%s)  ...", JSON.stringify(snmpDiscoveryData));
        for (var deviceMac in snmpDiscoveryData) {
            if (snmpDiscoveryData.hasOwnProperty(deviceMac)) {
                var payload = {'status': snmpDiscoveryData[deviceMac].status};
                engineRequest.cachedSendWithPromise('/device/' + deviceMac + '/check-status', 'PUT', payload);
            }
        }
        return snmpDiscoveryData;
    };

    ResultHandler.prototype.triggerCascadingDiscoveries = function (snmpDiscoveryData) {
        myConsole.info("Triggering cascading discoveries for data %s ..", JSON.stringify(snmpDiscoveryData));
        var mustTriggerCascadingDiscoveries = false;
        Object.keys(snmpDiscoveryData).forEach(function (key) {
            if (snmpDiscoveryData[key].status === 'ON') {
                mustTriggerCascadingDiscoveries = true;
            }
        });
        myConsole.info('Should discoveries be forced %s ', mustTriggerCascadingDiscoveries);

        if (mustTriggerCascadingDiscoveries) {
            myConsole.info('Forcing discoveries ...');

            mibDiscoveryEventManager.discoveryForcing();
        }
        return snmpDiscoveryData;
    };

    var discover = function (address, mac, callback) {
        communitiesProvider.get().then(function (communities) {
            var snmpCommand = commandBuilder.buildBulkGetCommand(SYS_DESCR_OID, address);
            myConsole.debug("Command for device %s [mac %s]  %s ", address, mac, JSON.stringify(snmpCommand));
            var task = new TaskBuilder().deferredSnmpBulkGetTask(snmpCommand, address, communities[mac], mac);
            return executor.execute([task], myConsole)
            .then(new ResultHandler().renderSnmpResults)
            .then(saveDeviceToInternalCache)
            .then(function (results) {
                myConsole.debug("Finally, results are %s", JSON.stringify(results));
                if (callback) {
                    callback(results);
                }
                return results;
            })
            .then(new ResultHandler().triggerCascadingDiscoveries)
            .catch(function (error) {
                myConsole.error("Error on snmp discover: " + error);
            });
        });
    };

    var saveDeviceToInternalCache = function (results) {
        Object.keys(results).forEach(function(key, indx) {
            snmpDiscoverySnapshotData[key] = results[key].status;
            myConsole.debug("Results saved to internal cache %s ", JSON.stringify(snmpDiscoverySnapshotData[key]));
        });
        return results;
    };

    var discoverAll = function (callback, errCallback) {
        if (!lockDiscovery()) {
            myConsole.warn('WARNING: SNMP Discovery already in progress, ignoring the call');
            return q(null);
        }

        return communitiesProvider.get().then(function (communities) {
            var deferredResultList = [];
            myConsole.debug('Found communities %s ', JSON.stringify(communities));

            interfacesBinding.listFresh().forEach(function (device) {
                var address = device.ip;
                var snmpCommand = commandBuilder.buildBulkGetCommand(SYS_DESCR_OID, address);
                myConsole.debug('Command for device %s [mac %s]  %s ', device.ip, device.mac, JSON.stringify(snmpCommand));
                deferredResultList.push(
                    new TaskBuilder().deferredSnmpBulkGetTask(snmpCommand, address, communities[device.mac], device.mac));
            });
            return executor.execute(deferredResultList, myConsole)
                .then(new ResultHandler().renderSnmpResults)
                .then(saveResultToInternalCache)
                .then(new ResultHandler().sendResultToEngine)
                .then(function () {
                    unlockDiscovery();
                    if (callback) {
                        callback(snmpDiscoverySnapshotData);
                    }
                });
        }).catch(function (err, err2) {
            unlockDiscovery();
            myConsole.error('Error on snmp information processing %s %s', err, err2);
            if (errCallback) {
                errCallback(err);
            }
        });
    };

    var saveResultToInternalCache = function (result) {
        snmpDiscoverySnapshotData = lodash.clone(result);
        return snmpDiscoverySnapshotData;
    };

    var getSnapshotData = function () {
        return lodash.clone(snmpDiscoverySnapshotData);
    };

    return {
        discover: discover,
        discoverAll: discoverAll,
        getSnapshotData: getSnapshotData,
        isDiscoveryInProgress: function () {
            return mutex.isLocked(DISCOVERY_SEMAPHORE);
        }
    };
};


module.exports.factory = snmpCheckStatusDiscoveryFactory;