/** This file is part of Domotz Agent.
 * Copyright (C) 2017  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Created by Alessio Sanfratello <asanfratello@domotz.com> on 09/02/17.
 */

const DISCOVERY_SEMAPHORE = 'SNMP_CUSTOM_SEMAPHORE';
const SNMP_DISCOVERY_TIMEOUT = 10000;
const snmpCommonDiscovery = require('./snmp_common');


var snmpDiscoveryFactory = function (resourceLocator, commandBuilder, responseRendering, snmpGetterFactory) {
    "use strict";

    var mutex = resourceLocator.utils.mutex;
    var q = resourceLocator.q;
    var engineRequest = resourceLocator.engineRequest;
    var snmpCommonTransformer = resourceLocator.snmpCommon;
    var snmpDiscoverySnapshotData = {};
    var macToOidList = null;
    var myConsole = resourceLocator.log.decorateLogs();
    var interfacesBinding = resourceLocator.interfacesBindingStorage;
    var snmpCommon = snmpCommonDiscovery.snmpCommonOptions();

    function invalidate() {
        macToOidList = null;
    }

    var lockDiscovery = function () {
        return mutex.lock(DISCOVERY_SEMAPHORE, 1);
    };

    var unlockDiscovery = function () {
        mutex.unlock(DISCOVERY_SEMAPHORE, 1);
    };


    var deferredSnmpBulkGet = function (snmpCommand, host, options, mac) {
        var deferred = q.defer();
        var commandOptions = snmpCommon.getSnmpOptions(options);
        myConsole.debug('Command Options are  %s -> %s', mac, JSON.stringify(commandOptions));
        snmpGetterFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                myConsole.debug('Received result for %s: %s Bytes', host, JSON.stringify(result).length);
                result.host = host;
                result.mac = mac;
                deferred.resolve(result);
            },
            commandOptions
        );

        return deferred.promise;

    };

    var renderDiscoveryData = function (rawResultList) {
        myConsole.debug('Rendering discover raw list : |%s|', JSON.stringify(rawResultList));
        rawResultList.forEach(function (rawResult) {
            var ipAddress = rawResult.host;
            var mac = rawResult.mac;
            var data = responseRendering(resourceLocator, rawResult);

            myConsole.debug('Parsed Result: %s - |%s|', ipAddress, JSON.stringify(data));

            if (data.error === false) {
                snmpDiscoverySnapshotData[mac] = data.oids;
            }
        });
    };

    var doDiscovery = function (callback) {
        snmpDiscoverySnapshotData = {};
        var communitiesProvider = resourceLocator.snmpCommunitiesProvider;
        return communitiesProvider.get().then(function (communities) {
            myConsole.debug("Found communities %s ", JSON.stringify(communities));

            var devicePromises = [];
            interfacesBinding.listFresh().forEach(function (device) {
                if (macToOidList[device.mac] === undefined) {
                    return;
                }
                myConsole.info("Gathering information for %s", device.ip);

                var snmpCommand = commandBuilder.buildBulkGetCommand(macToOidList[device.mac], device.ip,
                    SNMP_DISCOVERY_TIMEOUT);
                snmpCommand.customTransformer = snmpCommonTransformer.type_invariant_transformer;
                devicePromises.push(deferredSnmpBulkGet(snmpCommand, device.ip, communities[device.mac], device.mac));
            });

            return q.all(devicePromises)
                .then(renderDiscoveryData)
                .then(function () {
                    unlockDiscovery();
                    if (callback) {
                        callback(snmpDiscoverySnapshotData);
                    }
                });
        });
    };

    var discover = function (callback, errCallback) {
        if (!lockDiscovery()) {
            myConsole.warn("WARNING: SNMP Custom Discovery already in progress, ignoring the call");
            return q('BUSY');
        }

        return getOidListToDiscover().then(function () {
            return doDiscovery(callback);
        }).then(function () {
            return 'FREE';
        }).catch(function (err, err2) {
            myConsole.error('Error performing discovery: %s %s', err, err2);
            errCallback(err, err2);
        }).finally(function () {
            unlockDiscovery();
            myConsole.debug("Discovery unlocked");
        });

    };


    var getOidListToDiscover = function () {
        if (macToOidList === null) {
            myConsole.debug('Requesting OID sensors to engine');
            return engineRequest.sendWithPromise('/oid-sensor', 'GET').then(function (response) {
                macToOidList = response;
                myConsole.debug("Counting %s oid sensors", Object.keys(macToOidList).length);
            });
        } else {
            myConsole.debug('Using OID sensors to inspect in memory');
            return q();
        }
    };


    /**
     * Send to engine all data gathered from devices with one HTTP post
     * @param discoveryData data gathered from devices
     */

    var sendDiscoveryDataToEngine = function (discoveryData) {
        var gotOIDS = Object.keys(discoveryData).length;
        if (gotOIDS === 0) {
            myConsole.info("No OID data gatherted, sending nothing to engine");
        } else {
            myConsole.info(" Records to send %d", gotOIDS);
            try {
                var resource = "/oid-sensor/sample";
                myConsole.info(" Sending data to engine");

                engineRequest.sendWithPromise(resource, 'POST', discoveryData, null, null,
                    {doNotFollow204: true})
                    .catch(function (e) {
                        myConsole.error("Error sending SNMP data to engine: %s", JSON.stringify(e));
                    });
            } catch (e) {
                myConsole.error("Error preparing data to send %s", JSON.stringify(e));
            }
        }
    };

    return {
        discover: discover,
        invalidate: invalidate,
        sendDiscoveryDataToEngine: sendDiscoveryDataToEngine,
        isDiscoveryInProgress: function () {
            return mutex.isLocked(DISCOVERY_SEMAPHORE);
        },
        getCurrentSnapshotData: function () {
            return snmpDiscoverySnapshotData;
        }
    };
};
module.exports.factory = snmpDiscoveryFactory;
