/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz UK LLP
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/
/**
 * Created by Massimiliano Cuzzoli <massi@domotz.com> on 18/10/16.

 */

const DISCOVERY_SEMAPHORE = 'SNMP_INTERFACE_SEMAPHORE';
const OIDS_TO_DISCOVER = {
    '1.3.6.1.2.1.2.2.1.3': 'int',         // ifType
    '1.3.6.1.2.1.31.1.1.1.1': 'string',   // ifName
    '1.3.6.1.2.1.2.2.1.2': 'string',      // ifDescription
    '1.3.6.1.2.1.31.1.1.1.15': 'int',     // ifHighSpeed
    '1.3.6.1.2.1.31.1.1.1.18': 'string',  // ifAlias
    '1.3.6.1.2.1.2.2.1.7': 'int',         // ifAdminStatus
    '1.3.6.1.2.1.2.2.1.8': 'int',         // ifOperationalStatus
    '1.3.6.1.2.1.2.2.1.10': 'int',        // ifInOctets
    '1.3.6.1.2.1.2.2.1.16': 'int',        // ifOutOctets
    '1.3.6.1.2.1.2.2.1.13': 'int',        // ifInDiscards
    '1.3.6.1.2.1.2.2.1.19': 'int',        // ifOutDiscards
    '1.3.6.1.2.1.2.2.1.14': 'int',        // ifInErrors
    '1.3.6.1.2.1.2.2.1.20': 'int'         // ifOutErrors
};
const SNMP_SUBTREE_TIMEOUT = 20000;
const TASK_DELAY = 20000;
const snmpCommonDiscovery = require('./snmp_common');
const executorStrategy = require('./snmp_sequential_executor');

var snmpDiscoveryFactory = function (resourceLocator, commandBuilder, responseRendering, snmpSubtreeFactory) {
    "use strict";
    var myConsole = resourceLocator.log.decorateLogs();
    var snmpInterfacesNumberStorage = resourceLocator.snmpInterfacesNumberStorage;
    var communitiesProvider = resourceLocator.snmpCommunitiesProvider;
    var interfacesBinding = resourceLocator.interfacesBindingStorage;
    var mutex = resourceLocator.utils.mutex;
    var q = resourceLocator.q;
    var engineRequest = resourceLocator.engineRequest;
    var snmpDiscoverySnapshotData = {};
    var snmpCommon = snmpCommonDiscovery.snmpCommonOptions();
    var platform = resourceLocator.utils.platform;
    var interfaceTypesToRender = resourceLocator.settingsManager.getInterfaceDiscoveryTypesSettings();

    var executor = platform.isSlow() ? executorStrategy.sequentialExecutor() : executorStrategy.parallelExecutor();

    var deferredSnmpSubtree = function (snmpCommand, host, options, mac) {
        var deferred = q.defer();
        var commandOptions = snmpCommon.getSnmpOptions(options);
        myConsole.debug('Command Options are  %s -> %s', mac, JSON.stringify(commandOptions));
        snmpSubtreeFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                myConsole.debug('Received result for %s', host);
                result.host = host;
                result.mac = mac;
                result.timestamp = new Date().toISOString();
                setTimeout(function () {
                    deferred.resolve(result);
                }, TASK_DELAY);
            },
            commandOptions
        );

        return deferred.promise;
    };

    var deferredSnmpSubtreeTask = function (snmpCommand, host, options, mac) {
        if (!platform.isSlow()) {
            return deferredSnmpSubtree(snmpCommand, host, options, mac);
        } else {
            return function task() {
                return deferredSnmpSubtree(snmpCommand, host, options, mac);
            };
        }
    };

    var lockDiscovery = function () {
        return mutex.lock(DISCOVERY_SEMAPHORE, 1);
    };

    var unlockDiscovery = function () {
        mutex.unlock(DISCOVERY_SEMAPHORE, 1);
    };

    var gatherInterfaceInfoForDevice = function (ipAddress, options, mac) {
        var snmpCommand = commandBuilder.buildSubtreeCommand(OIDS_TO_DISCOVER, ipAddress, SNMP_SUBTREE_TIMEOUT);
        myConsole.info("Gathering information for %s (%s)", ipAddress, mac);
        return deferredSnmpSubtreeTask(snmpCommand, ipAddress, options, mac);
    };

    var renderDiscoveryData = function (rawResultList) {
        for (var i = 0; i < rawResultList.length; i++) {
            var dataToRender = rawResultList[i];
            myConsole.debug("Rendering discovery data for %s", dataToRender.host);
            snmpDiscoverySnapshotData[dataToRender.mac] = responseRendering(dataToRender, interfaceTypesToRender);
        }
    };

    var discover = function (callback, errCallback) {

        if (!lockDiscovery()) {
            myConsole.warn("Already in progress, ignoring the call");
            return;
        }

        return communitiesProvider.get().then(function (communities) {
            var deferredResultList = [];
            myConsole.debug("Found communities %s ", JSON.stringify(communities));

            interfacesBinding.listFresh().forEach(function (device) {
                var interfaces = snmpInterfacesNumberStorage.fetch(device.mac);
                myConsole.debug("Device %s has %s interfaces", device.mac, interfaces);
                if (interfaces > 0) {
                    myConsole.debug("Communities for device  [mac %s]  %s ", device.mac, JSON.stringify(communities[device.mac]));
                    deferredResultList.push(gatherInterfaceInfoForDevice(device.ip, communities[device.mac], device.mac));
                }
            });
            return executor.execute(deferredResultList, myConsole)
                .then(function (results) {
                    renderDiscoveryData(results);
                    unlockDiscovery();
                    if (callback) {
                        callback(snmpDiscoverySnapshotData);
                    }
                })
                .catch(function (err) {
                    unlockDiscovery();
                    myConsole.error("Error on snmp information processing", err);
                    if (errCallback) {
                        errCallback(err);
                    }
                });

        });
    };

    var sendDiscoveryDataToEngine = function (discoveryData) {
        myConsole.info("Number of records to send %d", Object.keys(discoveryData).length);
        try {
            var logError = function (e) {
                myConsole.error("Error sending snmp data to Engine: %s", JSON.stringify(e));
            };
            for (var mac in discoveryData) {
                if (discoveryData.hasOwnProperty(mac) && discoveryData[mac] && discoveryData[mac].oids) {
                    myConsole.info("Sending data for device MAC=%d", mac);
                    engineRequest.sendWithPromise('/subnet/ip/device/' + mac + '/interface/snmp-info', 'PUT',
                        discoveryData[mac]).catch(logError);
                }
            }
        } catch (e) {
            myConsole.error("Error preparing data to send %s", JSON.stringify(e));
        }
    };

    return {
        discover: discover,
        sendDiscoveryDataToEngine: sendDiscoveryDataToEngine,
        isDiscoveryInProgress: function () {
            return mutex.isLocked(DISCOVERY_SEMAPHORE);
        },
        getCurrentSnapshotData: function () {
            return snmpDiscoverySnapshotData;
        }
    };
};


module.exports.factory = snmpDiscoveryFactory;
