/** This file is part of Domotz Agent.
 * Copyright (C) 2021  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Created by Fausto Lelli <fausto@domotz.com> 
 */ 

const DISCOVERY_SEMAPHORE = 'SNMP_OVERLAY_SEMAPHORE';
const SNMP_GET_TIMEOUT = 15000;
const SNMP_SUBTREE_TIMEOUT = 20000;
const snmpCommonDiscovery = require('./snmp_common');


var snmpOverlayDiscoveryFactory = function (resourceLocator, snmpGetterFactory, snmpSubtreeFactory) {
    'use strict';

    var mutex = resourceLocator.utils.mutex;
    var q = resourceLocator.q;
    var lodash = resourceLocator.lodash;

    var snmpOverlayDb = resourceLocator.discovery.snmp_overlay_db;
    var snmpOverlayDeviceDb = resourceLocator.discovery.snmp_overlay_device_db;
    var myConsole = resourceLocator.log.decorateLogs("SNMP_OVERLAY_DISCOVERY");
    var interfacesBinding = resourceLocator.interfacesBindingStorage;
    var snmpCommon = snmpCommonDiscovery.snmpCommonOptions();
    var engineRequest = resourceLocator.engineRequest;
    var commandBuilder = require('./command_builder').factory(resourceLocator);


    var lockDiscovery = function () {
        return mutex.lock(DISCOVERY_SEMAPHORE, 1);
    };

    var unlockDiscovery = function () {
        mutex.unlock(DISCOVERY_SEMAPHORE, 1);
    };

    var deferredSnmpSubtree = function (ip, mac, options, overlayId, oids) {
        var deferred = q.defer();
        var snmpCommand = commandBuilder.buildSubtreeCommand(oids, ip, SNMP_SUBTREE_TIMEOUT);
        var commandOptions = snmpCommon.getSnmpOptions(options);
        snmpSubtreeFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                if (myConsole.isEnabledFor('debug')) {
                    myConsole.debug('Received result for SUBTREE (mac %s) (ip %s) (result %s)', mac, ip, JSON.stringify(result));
                }
                deferred.resolve({
                    device_hw_address: mac,
                    overlay_id: overlayId,
                    oid_column: result
                });
            },
            commandOptions
        );

        return deferred.promise;
    };

    var deferredSnmpBulkGet = function (ip, mac, options, overlayId, oids) {
        var deferred = q.defer();
        var snmpCommand = commandBuilder.buildBulkGetCommand(oids, ip, SNMP_GET_TIMEOUT);
        var commandOptions = snmpCommon.getSnmpOptions(options);
        snmpGetterFactory(snmpCommand, resourceLocator).execute(
            function (result) {
                if (myConsole.isEnabledFor('debug')) {
                    myConsole.debug('Received result for GET (mac %s) (ip %s) (result %s)', mac, ip, JSON.stringify(result));
                }
                deferred.resolve({
                    device_hw_address: mac,
                    overlay_id: overlayId,
                    oid_scalar: result
                });
            },
            commandOptions
        );

        return deferred.promise;

    };

    var renderDiscoveryData = function (rawResultList) {
        if (myConsole.isEnabledFor('debug')) {
            myConsole.debug('Rendering discover raw list : |%s|', JSON.stringify(rawResultList));
        }
        sendDiscoveryDataToCloud(rawResultList);
        return rawResultList;
    };

    function getAllDevicesFreshMap() {
        var toReturn = {};
        var devices = interfacesBinding.getLastSnapshot();
        for (var i = 0; i < devices.length; i++) {
            toReturn[devices[i].mac] = devices[i].ip;
        }
        return toReturn;
    }

    var doDiscovery = function (payload, callback) {
        var overlayDeviceDb = [];
        if (payload) {
            if (myConsole.isEnabledFor('debug')) {
                myConsole.debug("PAYLOAD %s", JSON.stringify(payload));
            }
            overlayDeviceDb = payload.device_overlay;
            if (myConsole.isEnabledFor('debug')) {
                myConsole.debug("Device Overlay from payload %s", JSON.stringify(overlayDeviceDb));
            }
        } else {
            overlayDeviceDb = snmpOverlayDeviceDb.getOverlayDeviceDB();
        }
        if (myConsole.isEnabledFor('debug')) {
            myConsole.debug('Overlay Device DB: %s', JSON.stringify(overlayDeviceDb));
        }
        if (overlayDeviceDb === undefined || overlayDeviceDb === null || overlayDeviceDb.length === 0) {
            myConsole.info('Overlay device DB empty, null or undefined. Skip discovery');
            return;
        }

        var communitiesProvider = resourceLocator.snmpCommunitiesProvider;
        communitiesProvider.get().then(function (communities) {
            var devicePromises = [];
            
            var devicesMap = getAllDevicesFreshMap();
            overlayDeviceDb.forEach(function (deviceOverlay) {
                var overlayId = deviceOverlay.overlay_id;
                var deviceMac = deviceOverlay.device_hw_address;
                var deviceIp = devicesMap[deviceMac];
                myConsole.info('Performing discovery: (mac %s) (ip %s) (overlay_id %s)', deviceMac, deviceIp, overlayId);
                if (deviceIp === undefined) {
                    myConsole.info('Can not perform discovery for device: %s', deviceMac);
                    return;
                }
                var overlay;
                if (payload) {
                    payload.overlays.forEach(function (item) {
                        if(item.id === overlayId) {
                            overlay = item;
                            if (myConsole.isEnabledFor('debug')) {
                                myConsole.debug("Overlay from payload %s", JSON.stringify(overlay));
                            }
                        }
                    });
                } else {
                    overlay = snmpOverlayDb.getOverlay(overlayId);
                }
                if (overlay === undefined) {
                    myConsole.info('Missing overlay info for overlay_id: %s', overlayId);
                    return;
                }
                if (overlay.oid_column !== undefined && Object.keys(overlay.oid_column).length > 0) {
                    myConsole.debug('Check SUBTREE (mac %s) (overlay_id %s) (oids %s)', deviceMac, overlayId, JSON.stringify(overlay.oid_column));
                    devicePromises.push(deferredSnmpSubtree(deviceIp, deviceMac, communities[deviceMac], overlayId, overlay.oid_column));
                }
                if (overlay.oid_scalar !== undefined && Object.keys(overlay.oid_scalar).length > 0) {
                    myConsole.debug('Check GET (mac %s) (overlay_id %s) (oids %s)', deviceMac, overlayId, JSON.stringify(overlay.oid_scalar));
                    devicePromises.push(deferredSnmpBulkGet(deviceIp, deviceMac, communities[deviceMac], overlayId, overlay.oid_scalar));
                }
            });

            q.all(devicePromises)
                .then(function (results) {
                    if (myConsole.isEnabledFor('debug')) {
                        myConsole.debug('overlay discovery final result %s', JSON.stringify(results));
                    }
                    if (callback) {                
                        myConsole.info('replying result through callback ...' );
                        callback(results);
                    } else {
                        myConsole.info('Uploading result to cloud ...' );
                        renderDiscoveryData(results);
                    }
                });
        });
    };

    var discover = function (payload, callback) {
        if (!lockDiscovery()) {
            myConsole.warn('WARNING: SNMP Overlay Discovery already in progress, ignoring the call');
            return;
        }

        try {
            doDiscovery(payload, callback);
        } finally {
            unlockDiscovery();
            myConsole.info('Discovery unlocked');
        }
    };


    /**
     * Send to engine all data gathered from devices with one HTTP post
     * @param discoveryData data gathered from devices
     */

    var sendDiscoveryDataToCloud = function (discoveryData) {

        myConsole.info('Overlay discovery FINISHED, Results number to upload %s', lodash.size(discoveryData));
        if( myConsole.isEnabledFor("debug")) {
            myConsole.debug('Overlay discovery FINISHED, Results to upload %s', JSON.stringify(discoveryData));
        }

        try {
            var resource = "/overlay/sample";
            engineRequest.sendWithPromise(resource, 'POST', {data: discoveryData}, null, null,
                { doNotFollow204: true })
                .catch(function (e) {
                    myConsole.error("Error sending Overlay data to cloud: %s", JSON.stringify(e));
                });
        } catch (e) {
            myConsole.error("Error sending data %s", JSON.stringify(e));
        }

    };

    return {
        discover: discover,
        sendDiscoveryDataToEngine: sendDiscoveryDataToCloud,
        isDiscoveryInProgress: function () {
            return mutex.isLocked(DISCOVERY_SEMAPHORE);
        },
    };
};
module.exports.factory = snmpOverlayDiscoveryFactory;
