/** This file is part of Domotz Agent.
 * Copyright (C) 2020  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Alessandro Pagiaro <apagiaro@domotz.com> on 10/02/19.
 *
 */

// To add options check ssh manual. We assume to use standard openssh client.
const OPTIONS_MAP = {
    cipher: '-c',
    ssh_config: '-F',
    limit: '-l',
    ssh_option: '-o',
    port: '-P',
    program: '-S',
    recursive: '-r',
    verbose: '-v',
};

const SCP_BINARY = 'scp';
const READ = 'read';
const LINUX_SSHPASS_BINARY = 'domotz_sshpass';
const WINDOWS_SCP_BINARY = 'domotz_pscp.exe';

module.exports.factory = function (procUtils, fileUtils, childProcess, q, platform) {

    /**
     * SCP is an external tool. It must be linked with system libraries!
     */
    function _getCleanEnv () {
        var cleanedEnv = Object.create(process.env);
        cleanedEnv.LD_LIBRARY_PATH = '';
        return cleanedEnv;
    }

    function getArgsForAuthentication (password) {
        var args = [];

        args.push(LINUX_SSHPASS_BINARY);
        args.push('-p');
        args.push(password);
        args.push(SCP_BINARY);
        var currentPlatform = platform.getFullPlatform();
        if (currentPlatform && currentPlatform.indexOf('ubuntu_core') > -1) {
            args.push('-S');
            args.push(process.env.SNAP + '/usr/bin/ssh');
        }

        return args;

    }

    function getArgsFromOptions (options) {
        var args = [];
        Object.keys(options || {})
            .forEach(function (opt) {
                if (OPTIONS_MAP.hasOwnProperty(opt)) {
                    var optionName = OPTIONS_MAP[opt],
                        optionValue = options[opt];
                    if (optionName.indexOf('-o') === 0) {
                        args.push(optionName + optionValue);
                    } else if (typeof optionValue !== 'boolean') {
                        args.push(optionName);
                        args.push(optionValue);
                    } else {
                        args.push(optionName);
                    }
                }
            });
        return args;
    }

    function _attacchEventOnOutputs (scpProcess) {
        var result = {};
        scpProcess.stderr.on('data', function (data) {
            console.warn('ERR DATA ' + data);
            result.stdErr += data;
        });

        scpProcess.stdout.on('data', function (data) {
            console.verbose('STDOUT DATA ' + data);
            result.stdOut += data;
        });

        scpProcess.on('error', function (err) {
            console.warn('scp error', JSON.stringify(err));
        });
        return result;
    }

    function _performRead (temporaryOutputFile, scp, args, options) {
        var deferred = q.defer();
        var scpProcess = childProcess.spawn(scp, args, options);


        var processOutputs = _attacchEventOnOutputs(scpProcess);

        scpProcess.on('close', function (code) {
            console.info('scp process killed');
            return fileUtils.readFile(temporaryOutputFile, null, true)
                .then(function (fileRead) {
                    deferred.resolve({
                        content: Buffer.from(fileRead).toString('base64'),
                        stdout: processOutputs.stdOut,
                        stderr: processOutputs.stdErr,
                        return_code: code,
                    });
                }).catch(function () {
                    deferred.resolve({
                        content: null,
                        stdout: processOutputs.stdOut,
                        stderr: processOutputs.stdErr,
                        return_code: code,
                    });
                }).finally(function () {
                    scpProcess = null;
                });
        });
        return deferred.promise;
    }

    function _performWrite (temporaryOutputFile, content, scp, args, options) {
        var buff = new Buffer(content, 'base64');

        return fileUtils.writeFile(temporaryOutputFile, buff, true)
            .then(function () {
                var deferred = q.defer();
                var scpProcess = childProcess.spawn(scp, args, options);

                var processOutputs = _attacchEventOnOutputs(scpProcess);

                scpProcess.on('close', function (code) {
                    console.info('Scp process killed');
                    scpProcess = null;

                    deferred.resolve({
                        stdout: processOutputs.stdOut,
                        stderr: processOutputs.stdErr,
                        return_code: code,
                    });
                });
                return deferred.promise;
            });
    }

    function _generateFileName () {
        var os = require('os');
        var path = require('path');
        return path.join(os.tmpdir(), 'scp_file') + Math.floor(Date.now() / 1000);
    }

    function getArgs (password, options) {
        var args;
        if (platform.isWindows()) {
            args = [];
            args.push(WINDOWS_SCP_BINARY);
            args.push('-pw');
            args.push(password);
            args.push('-batch');
            args.push('-scp');
            args.push('-q');
            args.push('-P');
            args.push('22');
        } else {
            args = getArgsForAuthentication(password);
            args = args.concat(getArgsFromOptions(options));
            args = args.concat(['-oStrictHostKeyChecking=no', '-oUserKnownHostsFile=/dev/null']);
        }
        return args;
    }

    /**
     * @param operation: 'read' or 'write'
     * @param user: user used to perform the operation
     * @param host: host used to perform the operation
     * @param file_path: file to recover/write
     * @param options:
     *      --> password
     *      --> cipher
     *      --> ssh_config
     *      --> identity_file
     *      --> limit
     *      --> ssh_option
     *      --> ... see OPTIONS_MAP
     * }
     */
    function exec (operation, user, host, filePath, password, options, content) {

        var args = getArgs(password, options);

        var OUTPUT_FILE = command._generateFileName();

        var baseExecutable = args.shift();

        options.env = command._getCleanEnv();

        if (operation === READ) {
            args.push(user + '@' + host + ':' + filePath);
            args.push(OUTPUT_FILE);
            return command._performRead(OUTPUT_FILE, baseExecutable, args, options);
        } else {
            args.push(OUTPUT_FILE);
            args.push(user + '@' + host + ':' + filePath);
            return command._performWrite(OUTPUT_FILE, content, baseExecutable, args, options);
        }

    }

    var command = {
        _performRead: _performRead,
        _performWrite: _performWrite,
        _generateFileName: _generateFileName,
        _getCleanEnv: _getCleanEnv,
        exec: exec,
    };

    return command;
};
