/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Tommaso Latini <tommaso@domotz.com> on 06/09/16.
 *
 */

const MODULE = 'SSH_COMMAND_EXECUTOR';
// To add options check ssh manual. We assume to use standard openssh client.
const OPTIONS_MAP = {
    bind_address: '-b',
    cipher_spec: '-c',
    escape_char: '-e',
    config_file: '-F',
    dynamic_port_forwarding: '-D',
    identity_file: '-i',
    pkcs11: '-I',
    local_port_forwarding: '-L',
    mac_spec: '-m',
    disable_remote_command: '-N',
    host_key_checking: '-oStrictHostKeyChecking=',
    user_known_host_file: '-oUserKnownHostsFile=',
    server_alive_interval: '-oServerAliveInterval=',
    server_alive_count_max: '-oServerAliveCountMax=',
    kex_algorithm: '-oKexAlgorithms=',
    ctl_cmd: '-O',
    port: '-p',
    remote_port_forwarding: '-R',
    ctl_path: '-S',
    disable_pseudo_tty_alloc: '-T'
};

const USER = 'domotz';
const NULL_PASSWORD = null;
const NULL_PRIVATE_KEY = null;
const HOST = 'localhost';
const TTL = 3600000; // 1 hour
const SSH_BINARY = 'ssh';
const SSHPASS_BINARY = 'domotz_sshpass';
const USER_KNOWN_HOST_FILE = OPTIONS_MAP.user_known_host_file + '/dev/null';
const HOST_KEY_CHECKING = OPTIONS_MAP.host_key_checking + 'no';



var os = require('os');
var path = require('path');

module.exports.factory = function (procUtils, fileUtils, q) {

    function describe(args, options) {
        // Removing the password arguments from the logs
        var logged_args = args.slice().splice(2);
        return logged_args.join(' ') + (options ? '. Options: ' + JSON.stringify(options) : '');
    }

    /**
     * SSH is an external tool. It must be linked with system libraries!
     */
    function getCleanEnv() {
        var cleanedEnv = Object.create(process.env);
        cleanedEnv.LD_LIBRARY_PATH = "";
        return cleanedEnv;
    }

    function getOpt(options, names, dflt) {
        for (var i = 0; i < names.length; i++) {
            var name = names[i];
            var o = options[name];
            if (typeof o !== 'undefined') {
                delete options[name];
                return o;
            }
        }
        return dflt;
    }

    function getArgsForAuthentication(password, privateKey, identityFile) {
        var args = [];

        if (privateKey !== NULL_PRIVATE_KEY) {
            console.debug("%s - Private Key is present. use identity file: %s", MODULE, identityFile)
            //0600 cannot be changed to 0o0600 on node 0.10
            if (!fileUtils.createFile(identityFile, privateKey, {mode: 0600, encoding: 'ascii'})) { /* eslint-disable-line */
                privateKey = NULL_PRIVATE_KEY;
            } else {
                args.push(SSH_BINARY);
                args.push('-i');
                if (process.env.DPLATFORM === 'win') {
                    args.push("'"+ identityFile+ "'");
                } else {
                    args.push(identityFile);
                }
            }
        }

        if (privateKey === NULL_PRIVATE_KEY && password !== NULL_PASSWORD) {
            console.debug("%s - Fallback in sshpass", MODULE);
            args.push(SSHPASS_BINARY);
            args.push('-p');
            args.push(password);
            args.push(SSH_BINARY);
        }

        return args;

    }

    function getArgsFromOptions(options) {
        var args = [];
        Object.keys(options).forEach(function (opt) {
            if (OPTIONS_MAP.hasOwnProperty(opt)) {
                var arg1 = OPTIONS_MAP[opt],
                    arg2 = options[opt];
                if (arg1.indexOf('-o') === 0) {
                    args.push(arg1 + arg2);
                } else if (typeof arg2 !== 'boolean') {
                    args.push(arg1);
                    args.push(arg2);
                } else {
                    args.push(arg1);
                }
                delete options[opt];
            }
        });
        return args;
    }

    /**
     * @param command: Command to run in ssh <OPTIONAL>
     * @param options:
     *      --> log_prefix: prefix for log [default SSH_COMMAND_EXECUTOR]
     *      --> user: user login for remote shell [default domotz]
     *      --> password: password to forward to ssh using domotz_sshpass
     *      --> host: target device [default localhost]
     *      --> ttl: lifetime of channel [default 1 hour]
     *      --> ... see OPTIONS_MAP
     * }
     * @param correlationId - unique message id for trace logging
     */
    function exec(command, options, correlationId) {

        var logId = getOpt(options, ['log_prefix'], MODULE + ' - ' + correlationId + ' - ');
        var userKnownHostFile = getOpt(options, 'user_known_host_file', USER_KNOWN_HOST_FILE);
        var hostKeyChecking = getOpt(options, 'host_key_checking', HOST_KEY_CHECKING);
        var host = getOpt(options, ['host', 'hostname'], HOST);
        var user = getOpt(options, ['user', 'username'], USER);
        var password = getOpt(options, ['password', 'pwd'], NULL_PASSWORD);
        var privateKey = getOpt(options, ['private_key'], NULL_PRIVATE_KEY);
        var identityFile = path.join(os.tmpdir(), '.' + user);

        var args = getArgsForAuthentication(password, privateKey, identityFile);

        if (args.length === 0) {
            console.error('%s - Missing Authentication Method', logId);
            return q.reject();
        }

        args = args.concat(getArgsFromOptions(options));

        args.push(userKnownHostFile);
        args.push(hostKeyChecking);
        args.push(user + '@'+ host);

        if (command) {
            args.push(command);
        }

        var ssh = args.shift();

        if (password && privateKey === NULL_PRIVATE_KEY) {
            options.description = describe(args, options);
        }

        options.env = getCleanEnv();
        options.timeout = null;

        setTimeout(function() {
            fileUtils.deleteFile(identityFile);
            console.debug('%s - Successfully deleted file [%s]', logId, identityFile);
        }, options.ttl || 60000);

        if (process.env.DPLATFORM === 'win') {
            options.shell = process.env.DOMOTZ_ROOT_DIR + '\\lib\\portable-git\\bin\\domotz_bash.exe';
        }
        return procUtils.spawn(ssh, args, options);
    }

    return {
        // For testing
        _describe: describe,
        _getCleanedEnv: getCleanEnv,
        _getOpt: getOpt,
        _getArgsForAuthentication: getArgsForAuthentication,
        _getArgsFromOptions: getArgsFromOptions,
        // -----------
        exec: exec
    };
};