/** This file is part of Domotz Agent.
 * Copyright (C) 2020  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent. If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Andrea Azzara <a.azzara@domotz.com> on 16/04/20.
 */


module.exports.server = function (resourceLocator) {

    var myConsole = resourceLocator.log.decorateLogs();
    var tftp = resourceLocator.tftp;
    var host = resourceLocator.discovery.host;

    var filenameIndex = 0;
    var closed = true;
    var content;

    function generateVirtualFilename () {
        filenameIndex += 1;
        return 'f' + filenameIndex;
    }

    function getIPAddress () {
        return host.getIP();
    }

    function returnAndReleaseOnError (cb, e) {
        myConsole.error(e.message + '\n' + e.stack);
        closed = true;
        cb({error: e.toString()});
    }


    return {
        start: function (options, callback) {

            if (!closed) {
                callback({error: 'a server is already active'});
                return;
            }

            var timeout = parseInt(options.timeout_ms) || 30000;
            var retries = 3;
            if (options.retries !== undefined) {
                retries = options.retries;
            }

            try {
                content = new Buffer(options.content, 'base64');
            } catch (e) {
                myConsole.error('Unable to parse base64 content');
                returnAndReleaseOnError(callback, 'Unable to parse base64 content');
                return;
            }

            var filename = options.filename || generateVirtualFilename();
            var port = options.port || 69;
            var ipAddress = options.ip || getIPAddress();

            var connections = [];

            var server = tftp.createServer({
                port: port,
                host: ipAddress,
                denyPUT: true,
                blockSize:options.block_size || 1468,
                windowSize:options.windows_size || 4,
                retries: retries,
                timeout: options.server_timeout_ms || 3000
            }, function () {
                //disable default listener
            });

            function onRequestClose() {
                //Remove the connection
                connections.splice(connections.indexOf(this), 1);
                if (closed && !connections.length) {
                    //The server and all the connections have been closed
                    myConsole.info('Server closed');
                }
            }

            function onRequest (req, tftpRes) {
                req.on('error', function (error) {
                    //Error from the request
                    myConsole.error(error.message + '\n' + error.stack);
                });

                //Save the connection
                connections.push(req);

                //The "close" event is fired when the internal socket closes, regardless
                //whether it is produced by an error, the socket closes naturally due to the
                //end of the transfer or the transfer has been aborted
                req.on('close', onRequestClose);

                if (req.file === filename) {
                    tftpRes.setSize(content.length);
                    tftpRes.end(content);
                    myConsole.info('Content %s sent %s bytes', filename, content.length);
                } else {
                    myConsole.warn('unknown file name Request: "%s" not equal to Existing: "%s"', req.file, filename);
                    return req.abort(tftp.ENOENT);
                }

            }

            function onListening () {
                var result = {
                    port: port,
                    host: ipAddress,
                    filename: filename,
                };
                myConsole.info('Server listening %s, timeout=%s', JSON.stringify(result), timeout);
                callback(result);
            }

            function onError (error) {
                //Errors from the main socket
                returnAndReleaseOnError(callback, error);
            }

            function onClose () {
                closed = true;

                if (!connections.length) {
                    return myConsole.info('Server closed');
                }
                //Abort all the current transfers
                for (var i = 0; i < connections.length; i++) {
                    myConsole.info('Connection ' + i + ' aborted');
                    connections[i].abort();
                }
            }

            function onTimeout() {
                if (!closed) {
                    myConsole.info('Timeout expired, closing the server');
                    server.close();
                }
            }

            server.on('listening', onListening);
            server.on('request', onRequest);
            server.on('error', onError);
            server.on('close', onClose);

            server.listen();
            closed = false;

            if (timeout !== 0) {
                setTimeout(onTimeout, timeout);
            }
        },
    };
};
