/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Andrea Azzara <a.azzara@domotz.com> on 23/08/2017.
 */

const DEFAULT_NMAP_TIMEOUT = 28;
const DEFAULT_OVERLAY_DB_FD_CYCLES = 720;
const DEFAULT_OVERLAY_DB_STARTUP_FD_CYCLES = 30;
const DEFAULT_OVERLAY_DEVICE_DB_FD_CYCLES = 720;
const DEFAULT_OVERLAY_DEVICE_DB_STARTUP_FD_CYCLES = 40;

function SettingsManager(resourceLocator) {
    "use strict";
    var fileUtils = resourceLocator.utils.file;
    var myConsole = resourceLocator.log.decorateLogs();
    var fileName = process.env.DOMOTZ_CONF_DIR + '/agent_settings.json';
    var configurationUpdater = resourceLocator.configurationUpdater;

    function updateCustomInterfaces(refreshBindingsList) {
        myConsole.info("Updating custom settings: %s", JSON.stringify(resourceLocator.configuration.settings));

        var updateHostInfo = function (ifaces) {
            resourceLocator.discovery.host.loadCustomInterfaces(ifaces);
            if (refreshBindingsList) {
                resourceLocator.interfacesBindingStorage.invalidateList();
            }
        };


        if (!resourceLocator.configuration.settings.external_subnets && !resourceLocator.configuration.settings.external_hosts) {
            myConsole.info("No external subnets nor hosts found");
            updateHostInfo({});
            return;
        }

        var interfaces = {};

        if (resourceLocator.configuration.settings.external_subnets) {
            myConsole.info("Updating external subnets: %s", JSON.stringify(resourceLocator.configuration.settings.external_subnets));

            resourceLocator.configuration.settings.external_subnets.forEach(function (i) {
                interfaces['ext' + i.id] = {
                    'custom': true,
                    'netmask': i.subnet_mask,
                    'ip': i.address,
                    'mac': null
                };
            });
        } else {
            myConsole.info("No external subnets found");
        }


        if (resourceLocator.configuration.settings.external_hosts) {
            myConsole.info("Updating external hosts: %s", JSON.stringify(resourceLocator.configuration.settings.external_hosts));

            resourceLocator.configuration.settings.external_hosts.forEach(function (h) {
                interfaces['ext' + h.id] = {
                    'custom': true,
                    'netmask': 32,
                    'ip': h.ip_address,
                    'mac': null,
                    'hostname': h.hostname,
                    'discovery_hw_address': h.hw_address
                };
            });
        } else {
            myConsole.info("No external host found");
        }
        updateHostInfo(interfaces);
    }

    function checkConfigurationVersion() {
        myConsole.debug("Checking required conf version");
        var requiredConfigVersion = resourceLocator.lodash.get(resourceLocator, 'configuration.settings.required_conf_version');
        if (!requiredConfigVersion) {
            myConsole.debug("No required conf version found, ignore");
            return;
        }
        var confVersion = resourceLocator.lodash.get(resourceLocator.configuration, 'conf_version', null);
        myConsole.debug("Required conf version found %s, current one is %s", requiredConfigVersion, confVersion);
        if (requiredConfigVersion !== confVersion) {
            myConsole.info("Required conf version is different from current one " + requiredConfigVersion);
            var baseURL = resourceLocator.configuration.engine.api_endpoints.agent || resourceLocator.configuration.engine.api_endpoints.hub;
            var confURL = baseURL + 'agent/' + resourceLocator.configuration.id;
            configurationUpdater.updateConfiguration(requiredConfigVersion, confURL);
        }
    }

    function loadSettings() {
        myConsole.info("Reading settings from cache: %s", fileName);
        var cachedSettings = fileUtils.readFile(fileName);
        if (cachedSettings) {
            resourceLocator.configuration.settings = JSON.parse(cachedSettings);
            myConsole.info("Found settings from cache: %s", JSON.stringify(resourceLocator.configuration.settings));
            updateCustomInterfaces();
        } else {
            myConsole.info("Settings not found in cache, requiring settings from server");
            fetchSettings().then(updateCustomInterfaces);
        }
    }

    function writeSettings(settings) {
        var fileContents = JSON.stringify(settings, null, '  ');
        myConsole.info("Writing agent settings: %s", fileName);
        fileUtils.writeFileSafe(fileName, fileContents);
    }

    function fetchSettings(refreshBindingsList, callback) {
        var engineRequest = resourceLocator.engineRequest;

        return engineRequest.sendWithPromise('/settings', 'GET', {}, null, false).then(function (result) {
            myConsole.info('Updating settings with ' + JSON.stringify(result));
            resourceLocator.configuration.settings = result;
            writeSettings(result);
            updateCustomInterfaces(refreshBindingsList);
            checkConfigurationVersion();
            if (callback) {
                callback();
            }
        }).catch(function (error) {
            myConsole.error(error.message + '\n' + error.stack);
        });
    }

    function isOnvifUnicastDiscoveryEnabled() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.onvif_unicast_discovery', false);
    }

    function isBroadcastDiscoveryDisabled() {
        if (resourceLocator.configuration.settings && resourceLocator.configuration.settings.broadcast_discovery !== undefined && resourceLocator.configuration.settings.broadcast_discovery === false) {
            return true;
        }
        return false;
    }

    function isLowThrottleMode() {
        if (resourceLocator.configuration.settings && resourceLocator.configuration.settings.http_call_throttle !== undefined && resourceLocator.configuration.settings.http_call_throttle === 'LOW') {
            return true;
        }
        return false;
    }

    function getEyesInterval() {
        if (resourceLocator.configuration.settings && resourceLocator.configuration.settings.eyes_hourly_frequency && resourceLocator.configuration.settings.eyes_hourly_frequency != 0) {
            return parseInt(120 / resourceLocator.configuration.settings.eyes_hourly_frequency);
        }
        return null;
    }

    function getMibDiscoveryInterval() {
        try {
            if (resourceLocator.configuration.settings && resourceLocator.configuration.settings.oid_discovery && resourceLocator.configuration.settings.oid_discovery.oid_scan_period_fast_discoveries) {
                return parseInt(resourceLocator.configuration.settings.oid_discovery.oid_scan_period_fast_discoveries);
            }
            return null;
        } catch(err) {
            myConsole.warn("Error reading configuration for mib discovery: %s", JSON.stringify(err));
            myConsole.warn("Configuration is: %s", JSON.stringify(resourceLocator.configuration.settings));
            return null;
        }
    }

    function getMibDiscoveryStartupInterval() {
        try {
            if (resourceLocator.configuration.settings && resourceLocator.configuration.settings.oid_discovery && resourceLocator.configuration.settings.oid_discovery.oid_scan_delay_fast_discoveries) {
                return parseInt(resourceLocator.configuration.settings.oid_discovery.oid_scan_delay_fast_discoveries);
            }
            return null;
        } catch(err) {
            myConsole.warn("Error reading configuration for mib discovery: %s", JSON.stringify(err));
            myConsole.warn("Configuration is: %s", JSON.stringify(resourceLocator.configuration.settings));
            return null;
        }
    }

    function getOverlayDBInterval() {
        return parseInt(resourceLocator.lodash.get(resourceLocator, 'configuration.settings.overlay.db.fd_cycles', DEFAULT_OVERLAY_DB_FD_CYCLES));
    }

    function getOverlayDBStartupInterval() {
        return parseInt(resourceLocator.lodash.get(resourceLocator, 'configuration.settings.overlay.db.startup_fd_cycles', DEFAULT_OVERLAY_DB_STARTUP_FD_CYCLES));
    }

    function getOverlayDeviceDBInterval() {
        return parseInt(resourceLocator.lodash.get(resourceLocator, 'configuration.settings.overlay.device_db.fd_cycles', DEFAULT_OVERLAY_DEVICE_DB_FD_CYCLES));
    }

    function getOverlayDeviceDBStartupInterval() {
        return parseInt(resourceLocator.lodash.get(resourceLocator, 'configuration.settings.overlay.device_db.startup_fd_cycles', DEFAULT_OVERLAY_DEVICE_DB_STARTUP_FD_CYCLES));
    }

    function getNMAPDiscoveryTimeout() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.nmap_discovery_timeout', DEFAULT_NMAP_TIMEOUT);
    }

    function getRTDConfiguration () {
        if (!resourceLocator.lodash.get(resourceLocator, 'configuration.settings.rtd_scan', true)) {
            return null;
        }
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.rtd_discovery');
    }

    function getURLforNetinfo () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.netinfo.url_for_netinfo', "http://ipecho.net/plain");
    }

    function getURLforNetinfo2 () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.netinfo.url_for_netinfo_2', "http://icanhazip.com/");
    }

    function getDiscoverySettings () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.discovery', {
            "max_ip_netmask": 22,
            "ports": [
                22,
                23,
                80,
                443,
                3000,
                3389,
                5000,
                5900,
                8000,
                8001,
                8080,
                8888
            ],
            "udp_ports": [
                161
            ],
            "udp_port_status_mapping": {
                "CLOSED|FILTERED": "CLOSED",
                "OPEN|FILTERED": "CLOSED"
            },
            "refresh": 30
        });
    }

    function getMTRRemoteHost () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.mtr.remote_host_for_mtr_computation', "8.8.8.8");
    }

    function getMTRPacketNumber () {
        return parseInt(resourceLocator.lodash.get(resourceLocator, 'configuration.settings.mtr.number_of_packets_for_mtr', 10));
    }

    function getOidsToDiscover() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.oids_to_discover', {
            '1.3.6.1.2.1.1.1.0': 'string',  // sysDescr
            '1.3.6.1.2.1.1.2.0': 'string',  // sysOID
            '1.3.6.1.2.1.1.4.0': 'string',  // sysContact
            '1.3.6.1.2.1.1.5.0': 'string',  // sysName
            '1.3.6.1.2.1.1.6.0': 'string',  // sysLocation
            '1.3.6.1.2.1.1.7.0': 'int',     // sysServices
            '1.3.6.1.2.1.2.1.0': 'int'      // ifNumber
        });
    }

    function isOsInspectionsEnabled() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.os_inspection.enabled', false);
    }

    function osInspectionsPollingInterval() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.os_inspection.candidate_devices_polling_interval_hours', 3);
    }

    function isExperimentEnabled(experiment) {
        var experiments = resourceLocator.lodash.get(resourceLocator, 'configuration.settings.experiments', []);
        return experiments.indexOf(experiment) !== -1;
    }

    function getNetworkInterfacesPolicies() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.interfaces_policy', {
            "policy": "deny",
            "rules": []
        });
    }

    function isSnmpBasicDiscoveryEnabled () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.snmp.discovery.basic.enabled', true);
    }

    function isSnmpCheckStatusDiscoveryEnabled () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.snmp.discovery.check_status.enabled', true);
    }

    function isSnmpCustomDiscoveryEnabled () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.snmp.discovery.custom.enabled', true);
    }

    function isSnmpInterfaceDiscoveryEnabled () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.snmp.discovery.interface.enabled', true);
    }

    function isSnmpMibDiscoveryEnabled () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.snmp.discovery.mib.enabled', true);
    }

    function isSnmpOverlayDiscoveryEnabled () {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.snmp.discovery.overlay.enabled', false);
    }

    function isBigNetworksEnabled() {
        var bigNetworks = resourceLocator.lodash.get(resourceLocator, 'configuration.settings.big_networks', {"enabled": false});
        return bigNetworks && bigNetworks.enabled === true;
    }

    function bigNetworksBitmask() {
        if (!isBigNetworksEnabled()) {
            return null;
        }
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.big_networks.netmask_bits', 16);

    }

    function getBigNetworksPerformanceParams() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.big_networks.performance_params', {
           t_value: 4,
           timeout: 10,
           subnetwork_count: 256,
           parallel_executions: 4,
           cool_down_before_restart: 30000, // seconds of cool down before a big scan and another
        });
    }

    function getStallWatcherSettings() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.stall_watcher', {
            enabled: true,
            timeout: 5
        });
    }

    function getInterfaceDiscoveryTypesSettings() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.discovery.interfaces.if_types', [
            0, // fallback for non-compliant devices
            6, // ethernetCsmacd(6)
            62, // fastEther(62)
            69, // fastEtherFX(69)
            117 // gigabitEthernet(117)
        ]
        );
    }

    function getUDPScanSettings() {
        return resourceLocator.lodash.get(resourceLocator, 'configuration.settings.udp_scan', {
            use_nmap_custom_scripts: true
        })
    }

    return {
        loadSettings: loadSettings,
        writeSettings: writeSettings,
        fetchSettings: fetchSettings,
        isOnvifUnicastDiscoveryEnabled: isOnvifUnicastDiscoveryEnabled,
        isBroadcastDiscoveryDisabled: isBroadcastDiscoveryDisabled,
        isLowThrottleMode: isLowThrottleMode,
        getEyesInterval: getEyesInterval,
        getNMAPDiscoveryTimeout: getNMAPDiscoveryTimeout,
        getRTDConfiguration: getRTDConfiguration,
        getMibDiscoveryInterval: getMibDiscoveryInterval,
        getMibDiscoveryStartupInterval: getMibDiscoveryStartupInterval,
        getURLforNetinfo: getURLforNetinfo,
        getURLforNetinfo2: getURLforNetinfo2,
        getDiscoverySettings: getDiscoverySettings,
        getMTRRemoteHost: getMTRRemoteHost,
        getMTRPacketNumber: getMTRPacketNumber,
        getOidsToDiscover: getOidsToDiscover,
        getOverlayDBInterval: getOverlayDBInterval,
        getOverlayDBStartupInterval: getOverlayDBStartupInterval,
        getOverlayDeviceDBInterval: getOverlayDeviceDBInterval,
        getOverlayDeviceDBStartupInterval: getOverlayDeviceDBStartupInterval,
        isExperimentEnabled: isExperimentEnabled,
        isOsInspectionsEnabled: isOsInspectionsEnabled,
        osInspectionsPollingInterval: osInspectionsPollingInterval,
        getNetworkInterfacesPolicies: getNetworkInterfacesPolicies,
        isSnmpBasicDiscoveryEnabled: isSnmpBasicDiscoveryEnabled,
        isSnmpCheckStatusDiscoveryEnabled: isSnmpCheckStatusDiscoveryEnabled,
        isSnmpCustomDiscoveryEnabled: isSnmpCustomDiscoveryEnabled,
        isSnmpInterfaceDiscoveryEnabled: isSnmpInterfaceDiscoveryEnabled,
        isSnmpMibDiscoveryEnabled: isSnmpMibDiscoveryEnabled,
        isSnmpOverlayDiscoveryEnabled: isSnmpOverlayDiscoveryEnabled,
        isBigNetworksEnabled:isBigNetworksEnabled,
        bigNetworksBitmask:bigNetworksBitmask,
        getBigNetworksPerformanceParams: getBigNetworksPerformanceParams,
        getStallWatcherSettings: getStallWatcherSettings,
        getInterfaceDiscoveryTypesSettings: getInterfaceDiscoveryTypesSettings,
        getUDPScanSettings: getUDPScanSettings,
    };

}

module.exports.settingsManager = SettingsManager;
